; LAML Midi.
; Parsing of midi file to XML-in-LAML.
; Unparsing XML-in-LAML midi files to binary midi.

; Uncomment for debugging purposes:
; (require (lib "errortrace.ss" "errortrace"))

; ===============================================================================================================
; Template stuff. Temporary.
; Emacs  (setq interactive-laml-mirror-library nil)

; ===============================================================================================================
; CONSTANTS

; A boolean variable that controls if the midi LAML software prints (a few) messages during processing
(define midi-reporting? #t)


; ===============================================================================================================
; Global variables 



(define global-ppqn 1920)
(define global-signature (list 4 4))

; Give info about drum sounds in NoteOn messages
(define info-about-drum-sounds? #t)

; A list of channels considered as drum channels
(define drum-channels '(9 10))

(define (drum-sound-channel? channel-number)
  (memq channel-number drum-channels))

; ===============================================================================================================
; LOADING SECTION

(load (string-append laml-dir "lib/" "file-read.scm"))

(lib-load "xml-in-laml/xml-in-laml.scm")
(load (string-append laml-dir "styles/xml-in-laml/midi/mirror/midi.scm"))

; Convenient if you load this software repeatedly into a Scheme prompt:
(set! xml-check-language-overlap? #f)


; ---------------------------------------------------------------------------------------------------------------

(define xml-check-attributes? #f)
(define xml-validate-contents? #f)
(define xml-link-checking #f)
(define xml-check-language-overlap? #f)

; -----------------------------------------------------------------------------
; Integer handling functions.

; Return n. If n > ma, return ma. If n < mi return mi.
(define (between mi ma n)
  (max (min n ma) mi))

(define (to-int x)
  (truncate (exact->inexact x)))


(define (as-int-string x)
  (cond ((number? x) (remove-decimals (as-string x)))
        (else (as-string x))))

(define (remove-decimals num-string)
  (let ((point-pos (find-in-string num-string #\.)))
    (if point-pos
        (if (= point-pos 0) 
            "0"
            (substring num-string 0 point-pos))
        num-string)))


; ===============================================================================================================
; Overall processing functions.

;; Read and parse midi file in midi-file-path and make a corresponding laml source file in laml-file-path. 
;; midi-file-path is assumed to be a full path to the midi file (including file extension).
;; laml-file-path is assumed to be a full path to the generated laml file (in a preexisting directory).
;; Saves the intermediate AST in the variable the-ast for programmatic access.
;; mode is either raw, absTime or deltaTime (a symbol).
;; raw: delta times with NoteOn and NoteOff.
;; absTime: absolute times with durations.
;; deltaTime: delta times with duration.
;; counter-transp is the amount of transposition done to compensate for the transposition made at the keyboard. Provides for recognizable note names. An integer.
;; track-separation can be used on format 0 midi file, not pad files. A boolean.
(define (midi-file-to-laml midi-file-path laml-file-path mode counter-transp track-separation)
  (midi-report "This program converts a binary midi file to an XML-in-LAML source file.")
  (let ((input-file-extension (file-name-extension midi-file-path))
        (transformed-ast (midi-file-to-laml-ast midi-file-path mode counter-transp track-separation)))
    (midi-report "LAML AST has been made. Now rendering AST to a XML-in-LAML source file...")
    (set! the-ast transformed-ast)
    
    (make-midi-laml-file transformed-ast laml-file-path)
  )
)


;; Read and parse midi file in midi-file-path and return the corresponding AST.
;; midi-file-path is assumed to be a full path to the midi file (including file extension).
;; mode is either raw, absTime or deltaTime (a symbol).
;; raw: delta times with NoteOn and NoteOff.
;; absTime: absolute times with durations.
;; deltaTime: delta times with duration.
;; counter-transp is the amount of transposition done to compensate for the transposition made at the keyboard. Provides for recognizable note names. An integer.
;; track-separation can be used on format 0 midi file, not pad files. A boolean.
(define (midi-file-to-laml-ast midi-file-path mode counter-transp track-separation)
  (set! counter-transposition counter-transp)
  (let ((input-file-extension (file-name-extension midi-file-path))
        (midi-str (read-text-file midi-file-path))) 
    (estimate-processing-time! (string-length midi-str))
    (midi-report "Parsing the binary midi file...")
    (let* ((lst (parse-midi-string midi-str))
           (non-midi-suffix-str (get-non-midi-suffix-string midi-str lst))
          )
      (midi-report "Midi parsing done. Now constructing and transforming the XML AST...")
      (let* ((ast (midi-file-as-list-to-ast lst mode non-midi-suffix-str))  ; various requested transformations done by midi-file-as-list-to-ast
             (ast-header (ast-subtree ast "MidiHeader"))
             (midi-format (as-number (ast-attribute ast-header 'format)))
             (transformed-ast 
                (cond ; Midi file format 0
                      ((and (= midi-format 0) (or (equal? (downcase-string input-file-extension) "mid") (equal? (downcase-string input-file-extension) "midi")))
                          (if track-separation
                              (track-separated-standard-midi-file-ast ast standard-track-separation-scheme)  
                              ast)
                      )

                      ; Midi file format 1.   Disregard track-separation
                      ((and (= midi-format 1) (or (equal? (downcase-string input-file-extension) "mid") (equal? (downcase-string input-file-extension) "midi")))
                          ast
                      )

                      ; Yamaha Tyros Multipad
                      ((and (equal? (downcase-string input-file-extension) "pad") (or (equal? (as-string mode) "absTime") (equal? (as-string mode) "deltaTime")))  
                          (standard-midi-file-ast-to-multipad-ast ast midi-file-path))
                      ((and (equal? (downcase-string input-file-extension) "pad") (not (equal? (as-string mode) "absTime"))) 
                          (laml-error "Can only deal with multiPads in absTime mode or deltaTime mode" ))

                      ; Yamaha Tyros Styles
                      ((and (= midi-format 0) 
                            (member (downcase-string input-file-extension) (list "sty" "pst" "psc" "sst" "prs" "bcs")))
                          ast
                      )

                      (else (laml-error "Unknown file extension of input file. I only know about mid, midi, and pad" (downcase-string input-file-extension)))))
            )
        transformed-ast))))



; Return the part of the binary midi-string which comes after the standard midi contents.
; This may, for instance, be the CASM part of a Yamaha style file.
; Returns a (binary) string.
(define (get-non-midi-suffix-string midi-string parsed-midi-list)
  (let* ((midi-string-length (string-length midi-string))
         (header-length (byte-length-of-header (header-of-smf parsed-midi-list)))
         (track-length-list (map byte-length-of-track  (tracks-of-smf parsed-midi-list)))
         (track-length-sum (accumulate-right + 0 track-length-list)))
    (substring midi-string (+ header-length track-length-sum) midi-string-length)))

(define (byte-length-of-header parsed-header)
  (+ 4  ; "MThd"
     4  ; 4 byte length, always the number 6
     6  ; Header data length
  ))

(define (byte-length-of-track parsed-track)
  (+ 4                                    ; "MTrk"
     4                                    ; 4 byte length
     (length-of-midi-track parsed-track)  ; track data length
  ))



;; Bulk Generation of LAML files from Midi files.
;; Make laml files of all files in midi-file-list (a list of proper file names with extension). 
;; Each file in midi-file-list is assume to reside input-dir (full path, ends in forward slash).
;; Place output files in output-dir (full path, ends in forward slash).
;; mode is either raw, absTime or deltaTime (a symbol).
;; raw: delta times with NoteOn and NoteOff.
;; absTime: absolute times with durations.
;; deltaTime: delta times with duration.
(define (midi-files-to-laml-files input-dir midi-file-list output-dir mode)
  (for-each
    (lambda (f)
       (display-message f)
       (midi-file-to-laml 
         (string-append input-dir f)
         (string-append output-dir (file-name-proper f) "." "midl")
         mode
         0
         #f)
       (display-message "") 
       )
    midi-file-list))

(define (midi-file? path)
  (let ((ext (file-name-extension path)))
    (or (equal? (downcase-string ext) "mid") (equal? (downcase-string ext) "midi"))))

(define (estimate-processing-time! byte-lgt-midi-str)
  (midi-report
    (string-append
      "Length of binary midi-file: " (as-string byte-lgt-midi-str) " bytes" "."))
  (midi-report
    (string-append
      "Estimated processing time: " (as-string (max (to-int (* (/ 2 3) (quotient byte-lgt-midi-str 1000))) 2)) " seconds.")))


; A global variable which references the latest AST processed.
(define the-ast #f)


; ===============================================================================================================
; XML-in-LAML action procedure.

(define (StandardMidiFile! ast)
 (let ((mode-string (ast-attribute (ast-subtree ast "MidiHeader") 'mode))
       (source-file-name (source-filename-without-extension))  ; file-name-proper is used to strip y in file names x.y.ext.  source-file-name becomes x.
      )
  (ensure-directory-existence! (startup-directory) "midi-temp") 
  (write-text-file
     (standard-midi-file-ast-to-bin ast)
     (in-startup-directory "midi-temp/" (string-append source-file-name "." "mid")))))

(define (TyrosMultiPad! ast)
 (let* ((mode-string (ast-attribute ast 'mode))                                           ; mode not used in this version.
        (image (ast-attribute ast 'image "S910"))
        (transformed-ast (multipad-ast-to-standard-midi-file-ast ast)))
  (ensure-directory-existence! (startup-directory) "midi-temp") 
  (write-text-file
     (standard-midi-file-ast-to-bin transformed-ast)
     (in-startup-directory "midi-temp/"
                           (string-append (source-filename-without-extension)    ; With possible middle name, such as "xx.S384"
                                           "." "mid"))                           ; we use -.mid to provide for playing of the pad
  )))



; ===============================================================================================================
; Processing reportation stuff

(define (midi-report . mes)
  (if midi-reporting? 
      (apply display-message mes)))


; ===============================================================================================================
; OVERALL STANDARD MIDI FILE PARSING.
; The result of this parsing is an intermediate midi list structure.


; ----------------------------------------------------------------------------------------------------------
;;; CONSTRUCTORS AND SELECTORS - RELATED TO THE INTERMEDIATE LIST STRUCTURE REPRESENTATIONS

;; Make a standard midi file list structure given a parsed header and parsed tracks.
(define (make-smf header tracks)
  (cons 'smf (cons header tracks)))

;; Select and return the header of parsed standard midi file list structure
(define header-of-smf (make-selector-function 2 "header-of-smf"))

;; Select and return the tracks of parsed standard midi file list structure
(define (tracks-of-smf smf)
  (cdr (cdr smf))) 



;; Make a midi header chuck given format, number-of-tracks and division 
(define (make-midi-header-chunck format number-of-tracks division)
  (list 'mthd (as-number format) (as-number number-of-tracks) division))

;; Select and return the format of a parsed midi file header chunck.
(define format-of-midi-header (make-selector-function 2 "format-of-midi-header"))

;; Select and return the number of tracks of a parsed midi file header chunck.
(define number-of-tracks-of-midi-header (make-selector-function 3 "number-of-tracks-of-midi-header"))

;; Select and return the division (alwas in ppqn - pulses per quater note) of a parsed midi file header chunck.
(define (ppqn-of-midi-header header)
  (let ((division (list-ref header 3)))
    (if (eq? (first division) 'ppqn)
        (second division)
        #f)))



;; Make a single midi track chuck. data is a list of track events, each constructed by make-track-event.
(define (make-midi-track-chunck length data)
  (list 'mtrk (as-number length) data))

;; Select and return the length of a parsed track
(define length-of-midi-track (make-selector-function 2 "length-of-midi-track"))

;; Select and return the data of a parsed track. A list of track-events, each as constructed by make-track-event.
(define data-of-midi-track (make-selector-function 3 "data-of-midi-track"))



(define (make-track-event delta-time midi-event)
  (list 'track-event delta-time midi-event))

(define delta-time-of-track-event (make-selector-function 2 "delta-time-of-track-event"))
(define midi-event-of-track-event (make-selector-function 3 "midi-event-of-track-event")) 



(define (make-midi-event status-byte rest-bytes)
  (list 'midi-event status-byte rest-bytes))

(define status-byte-of-midi-event (make-selector-function 2 "status-byte-of-midi-event"))
(define rest-bytes-of-midi-event (make-selector-function 3 "rest-bytes-of-midi-event"))

; ---------------------------------------------------------------------------------------------------------

;; Parse a midi string return a smf list structure.
(define (parse-midi-string midi-string)
  (let* ((header (parse-header-of-midi-string midi-string))
         (number-of-tracks (number-of-tracks-of-midi-header header)))
    (make-smf header (parse-tracks-of-midi-string midi-string 14 number-of-tracks))))

;; Parse a header of standard midi file, as represented by the byte string str.
(define (parse-header-of-midi-string str)
  (if (and (equal? (substring str 0 4) "MThd")                   ; Byte 0-3
           (= (byte-string-to-integer (substring str 4 8)) 6))   ; Byte 4-7
      (make-midi-header-chunck 
         (byte-string-to-integer (substring str 8 10))           ; Byte 8-9
         (byte-string-to-integer (substring str 10 12))          ; Byte 10-11
         (parse-division (substring str 12 14))                  ; Byte 12-13
      )
      #f))

; Temporary - to find out where midi part ends
(define length-info '())

;; Parse a single track from string which starts at start-pos.
;; str is the overall byte-string that represents the whole midi file.
(define (parse-track-of-midi-string str start-pos)
  (if (equal? (substring str start-pos (+ start-pos 4)) "MTrk")
      (let ((track-length (byte-string-to-integer (substring str (+ start-pos 4) (+ start-pos 8)))))
        (set! length-info (cons (+ start-pos 8 track-length) length-info))
        (make-midi-track-chunck 
         track-length 
         (parse-track-data (substring str (+ start-pos 8) (+ start-pos 8 track-length)))
         ))
      #f))

;; Parse n tracks from str. The first one starts at position start-pos.
;; Return a list of tracks.
;; str is the overall byte-string that representins the whole midi file.
(define (parse-tracks-of-midi-string str start-pos n)
  (if (= n 0)
      '()
      (let* ((track (parse-track-of-midi-string str start-pos))
             (track-length (length-of-midi-track track)))
        (cons track (parse-tracks-of-midi-string str (+ start-pos 8 track-length) (- n 1))))))

;; Parse a division from the header chunck.
(define (parse-division two-byte-string)
  (if (>= (as-number (string-ref two-byte-string 0)) 128)        ; "first byte negative"
      (error "parse-division: only PPQN is supported as of now")
      (list 'ppqn (as-number (byte-string-to-integer two-byte-string)))))

; ---------------------------------------------------------------------------------------------------------------
; Track data parsing.

;; Parse track data in string. Str contains all data pertaining to the track. Nothing more and nothing less.
;; Returns a list of track events, each constructed by the function make-track-event.
(define (parse-track-data str)
  (parse-track-data-1 str (string-length str) 0 #f))

(define (parse-track-data-1 str str-lgt start-pos running-status-byte)
  (if (>= start-pos str-lgt)
      '()
      (let* ((track-event-data (parse-track-event str str-lgt start-pos running-status-byte))
             (track-event-length (car track-event-data))
             (the-track-event (cdr track-event-data))
             (new-running-status-byte (status-byte-of-midi-event (midi-event-of-track-event the-track-event)))
            )
        (cons
          the-track-event
          (parse-track-data-1 str str-lgt (+ start-pos track-event-length) new-running-status-byte)))))

; Parse a track event (delta time and midi event).
; Return cons cell of the byte length of the track event and and the track event as such.
; running-status-byte is the status byte (if available) of the previous midi event.
(define (parse-track-event str str-lgt start-pos running-status-byte)
;  (midi-report "parse-track-event" start-pos)
  (let* ((vlq-info (parse-variable-length-quantity str start-pos))   ; First delta time, as variable length quantity:  vlq
         (vlq-length (car vlq-info))
         (vlq (cdr vlq-info))
        )
    (let* ((midi-event-info (parse-midi-event str (+ start-pos vlq-length) running-status-byte))    ; Next Midi event: midi-event
           (midi-length (car midi-event-info))
           (midi-event (cdr midi-event-info))
          )
      (cons (+ vlq-length midi-length) (make-track-event vlq midi-event)))))


; Parse a midi event, starting at position start-pos.
; Runnning-status-byte is a byte, or #f if not available.
; Return a cons-cell of the length of the midi-event and the midi-event itself, as constructed by the function make-midi-event.
(define (parse-midi-event str start-pos running-status-byte)
  (let* ((first-byte-of-event (as-number (string-ref str start-pos)))
         (status-byte-of-midi-event (if (is-status-byte? first-byte-of-event) first-byte-of-event running-status-byte))
         (data-bytes-of-midi-event  (read-data-bytes-of-midi-event 
                                       str
                                       (if (is-status-byte? first-byte-of-event) (+ start-pos 1) start-pos) 
                                       status-byte-of-midi-event))
        )
    ; (if (not (is-status-byte? first-byte-of-event)) (midi-report "running status used"))
    (let ((lgt-data-bytes (string-length data-bytes-of-midi-event)))
     (cons 
      (if (is-status-byte? first-byte-of-event) (+ 1 lgt-data-bytes) lgt-data-bytes)
      (make-midi-event status-byte-of-midi-event data-bytes-of-midi-event)
     )
    )
  )
)

; Read the data bytes of a midi event. The data bytes start at start-pos.
; status-byte describes the type of the data-bytes to be parsed, and the amount of reading is done according to data-byte.
; Return a string of data bytes (everything after the status byte).
(define (read-data-bytes-of-midi-event str start-pos status-byte)
                                                                                                    ; STATUS BYTE
  (cond ((note-off-event? status-byte)                (substring str start-pos (+ start-pos 2)))    ; 1000 cccc  + two data bytes (the 4 cccc bits encode the channel)
        ((note-on-event? status-byte)                 (substring str start-pos (+ start-pos 2)))    ; 1001 cccc  + two data bytes
        ((polyphonic-aftertouch-event? status-byte)   (substring str start-pos (+ start-pos 2)))    ; 1010 cccc  + two data bytes
        ((control-change-event? status-byte)          (substring str start-pos (+ start-pos 2)))    ; 1011 cccc  + two data bytes
        ((program-change-event? status-byte)          (substring str start-pos (+ start-pos 1)))    ; 1100 cccc  + one data byte
        ((channel-pressure-event? status-byte)        (substring str start-pos (+ start-pos 1)))    ; 1101 cccc  + one data byte
        ((pitch-wheel-event? status-byte)             (substring str start-pos (+ start-pos 2)))    ; 1110 cccc  + two data bytes
        (else                                                                                       ; 1111 ****  + a variable number of bytes, where * is one bit.
                                                                                                    ; or in hex:   F*  where * is four bits.

                                                         ; Now notating in hexadecimal notation:
          (cond
            ((system-exclusive-event? status-byte)       ; F0 <vlq-length> <bytes> F7       Variable length data bytes.
                                                         ; F7 <vlq-length> <bytes> F7       Relies on trailing F7
               (let* ((f7-char (make-char-2-hex 15 7))   ; F = 15 (decimal).
                      (f7-end-pos (find-in-string str f7-char (+ start-pos 1))))
                 (substring str start-pos (+ f7-end-pos 1)))   ; including ending f7-char ! 

            )
            ((mtc-quater-frame-message-event? status-byte)   (substring str start-pos (+ start-pos 1)))   ; 1111 0001    or in Hex:  F1  +  One data byte.
            ((song-position-pointer-event? status-byte)      (substring str start-pos (+ start-pos 2)))   ; 1111 0010    or in Hex:  F2  +  Two data bytes.
            ((song-select-event? status-byte)                (substring str start-pos (+ start-pos 1)))   ; 1111 0011    or in Hex:  F3  +  One data byte.
            ((tune-request-event? status-byte)               "")                                          ; 1111 0110    or in Hex:  F6  +  NO data.
            ((midi-clock-event? status-byte)                 "")                                          ; 1111 1000    or in Hex:  F8  +  NO data.            
            ((midi-tick-event? status-byte)                  "")                                          ; 1111 1001    or in Hex:  F9  +  NO data.            
            ((midi-start-event? status-byte)                 "")                                          ; 1111 1010    or in Hex:  FA  +  NO data.            
            ((midi-continue-event? status-byte)              "")                                          ; 1111 1011    or in Hex:  FB  +  NO data.            
            ((midi-stop-event? status-byte)                  "")                                          ; 1111 1100    or in Hex:  FC  +  NO data.            
            ((midi-active-sense? status-byte)                "")                                          ; 1111 1110    or in Hex:  FE  +  NO data.            

            ; Midi reset - will not happen. Will never be sent. Therefore interpreted as midi-meta event:
            ((meta-event? status-byte)                       ; FF <extra-status-byte> <vlq-length> <bytes>
                                                             ; thus variable length data bytes.
               (let* ((vlq-info (parse-variable-length-quantity str (+ start-pos 1)))
                      (vlq-length (car vlq-info))
                      (vlq (cdr vlq-info)))
                  (substring str start-pos (+ start-pos 1 vlq-length vlq)))
            )
            (else (error (string-append "Non-existing system common, real-time and system exclusive event. Status-byte: " (as-string status-byte))))
          )
        )
  )
)

; ---------------------------------------------------------------------------------------------------------------
; Status byte predicates.

(define (note-off-event? status-byte)                ; 1000 cccc   (the 4 cccc bits encode the channel) 
  (and (>= status-byte 128) (<= status-byte 143)))

(define (note-on-event? status-byte)                 ; 1001 cccc
  (and (>= status-byte 144) (<= status-byte 159)))

(define (polyphonic-aftertouch-event? status-byte)   ; 1010 cccc
  (and (>= status-byte 160) (<= status-byte 175)))

(define (control-change-event? status-byte)          ; 1011 cccc
  (and (>= status-byte 176) (<= status-byte 191)))

(define (program-change-event? status-byte)          ; 1100 cccc
  (and (>= status-byte 192) (<= status-byte 207)))

(define (channel-pressure-event? status-byte)        ; 1101 cccc
  (and (>= status-byte 208) (<= status-byte 223)))

(define (pitch-wheel-event? status-byte)             ; 1110 cccc
  (and (>= status-byte 224) (<= status-byte 239)))

(define (system-exclusive-event? status-byte)        ; 1111 ...  ???
  (or (= status-byte 240) (= status-byte 247)))

(define (mtc-quater-frame-message-event? status-byte)
  (= status-byte 241))

(define (song-position-pointer-event? status-byte)
  (= status-byte 242))

(define (song-select-event? status-byte)
  (= status-byte 243))

(define (tune-request-event? status-byte)
  (= status-byte 246))

(define (midi-clock-event? status-byte)
  (= status-byte 248))

(define (midi-tick-event? status-byte)
  (= status-byte 249))

(define (midi-start-event? status-byte)
  (= status-byte 250))

(define (midi-continue-event? status-byte)
  (= status-byte 251))

(define (midi-stop-event? status-byte)
  (= status-byte 252))

(define (midi-active-sense? status-byte)
  (= status-byte 254))

(define (meta-event? status-byte)
  (= status-byte 255))

; --------------------------------------------------------------------------------
; Other byte predicates.

(define (is-status-byte? byte)
  (>= byte 128))


; ===============================================================================================================
; VARIABLE LENGTH QUANTITIES. PARSING AND UNPARSING.
; Variable length quantities, according to   http://www.borg.com/~jglatt/tech/midifile/vari.htm
; Variable length quantities are variable length, binary representation of integers.
; The first (most significant) bit in each byte encodes status: 1 means not last byte. 0 means last byte.
; Most significant byte at start-pos.
; Return a cons cell of the form (n . val) where n is the number of bytes processed, 
; and val is the integer value aggregated from these bytes.

(define (parse-variable-length-quantity str start-pos)
 (let* ((str-lgt (string-length str))
        (byte-1 (as-number (string-ref str start-pos)))
        (stop-1 (< byte-1 128))
        (byte-2 (if (< (+ start-pos 1) str-lgt) (as-number (string-ref str (+ start-pos 1))) #f))
        (stop-2 (if byte-2 (< byte-2 128) #t))
        (byte-3 (if (< (+ start-pos 2) str-lgt) (as-number (string-ref str (+ start-pos 2))) #f))
        (stop-3 (if byte-3 (< byte-3 128) #t))
        (byte-4 (if (< (+ start-pos 3) str-lgt) (as-number (string-ref str (+ start-pos 3))) #f))
        (stop-4 (if byte-4 (< byte-4 128) #t))
       )
   (cond ((and stop-1 byte-1) (cons 1 byte-1))
         ((and stop-2 byte-1 byte-2) (cons 2 (+ (* (- byte-1 128) 128) byte-2)))
         ((and stop-3 byte-1 byte-2 byte-3) (cons 3 (+ (* (- byte-1 128) 128 128) (* (- byte-2 128) 128) byte-3))) 
         ((and stop-4 byte-1 byte-2 byte-3 byte-4) (cons 4 (+ (* (- byte-1 128) 128 128 128) (* (- byte-2 128) 128 128) (* (- byte-3 128) 128) byte-4))) 
         (else (error "parse-variable-length-quantity: str does not encode a valid variable length quantity. At most four bytes are allowed")))))

; The inverse function of parse-variable-length-quantity.
; Return a byte string which represents n as a variable length quantiy.
; n is non-negative.
(define (unparse-integer-as-variable-length-quantity n)
  (let ((low (remainder n 128))
        (rest (quotient n 128)))
    (string-append (unparse-integer-as-variable-length-quantity-highs rest) (as-string (as-char low)))))

;; Produce bytes in which the high bits are all set to 1.  
(define (unparse-integer-as-variable-length-quantity-highs n)
  (if (= n 0)
      ""
      (let ((low (remainder n 128))
            (rest (quotient n 128)))
        (string-append (unparse-integer-as-variable-length-quantity-highs rest) (as-string (as-char (+ 128 low)))) ;    + 128  enforces one as first bit.
        )))



; ===============================================================================================================
; TEXTUAL REPORTING FUNCTIONS.
; Obsolete. Only used during development.

(define (present-midi-file file-path)
  (let ((parsed-midi-structure (parse-midi-string (read-text-file file-path))))
    (present-smf parsed-midi-structure)))

(define (present-smf smf-structure)
  (let ((header (header-of-smf smf-structure))
        (tracks (tracks-of-smf smf-structure)))
    (present-midi-header header)
    (display (string-append CR))
    (display (string-append  " -------------------------" CR))
    (for-each (lambda (n dtr)
                (display (string-append (as-string n)  " -------------------------" CR)) 
                (present-track-data (data-of-midi-track dtr)))
              (number-interval 1 (length tracks))
              tracks)))

(define (present-midi-header header)
  (display (string-append 
              "HEADER " CR
              "Format: " (as-string (format-of-midi-header header)) CR
              "Number of tracks: " (as-string (number-of-tracks-of-midi-header header)) CR
              "Pulses per quater note: " (as-string (ppqn-of-midi-header header)))))


(define (present-track-data track-event-list)
  (for-each
    (lambda (track-event)
       (display (string-append (track-event-as-string track-event) CR)))
    track-event-list))

; Produce a pretty string from a track event.
(define (track-event-as-string track-event)
  (string-append "TRE: " (as-string (delta-time-of-track-event track-event)) " "
                (midi-event-as-string (midi-event-of-track-event track-event))))

(define (midi-event-as-string midi-event)
  (let ((status-byte (status-byte-of-midi-event midi-event))
        (rest-bytes (rest-bytes-of-midi-event midi-event)))
    (present-midi-event status-byte rest-bytes)))

(define (present-midi-event status-byte rest-bytes)
  (cond ((note-off-event? status-byte)   (string-append "Note-off #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes))  )
        ((note-on-event? status-byte)    (string-append "Note-on #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes))  )
        ((polyphonic-aftertouch-event? status-byte) (string-append "Poly Aftertch #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes)) ) 
        ((control-change-event? status-byte)      (string-append "Control change #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes))   ) 
        ((program-change-event? status-byte)      (string-append "Program change #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes))   )
        ((channel-pressure-event? status-byte)    (string-append "Channel pressure #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes))   )
        ((pitch-wheel-event? status-byte)         (string-append "Pitch wheel #" (as-string (channel-of-status-byte status-byte)) " " (present-rest-bytes rest-bytes))   )   
        (else                                                                             
          (cond
            ((system-exclusive-event? status-byte)     
                (string-append "System Exclusive " (as-string (string-length rest-bytes))) 

            )
            ((mtc-quater-frame-message-event? status-byte) "mtc-quarter" )  
            ((song-position-pointer-event? status-byte)    "song pos" )   
            ((song-select-event? status-byte)              "song select" )
            ((tune-request-event? status-byte)             "tune req" )   
            ((midi-clock-event? status-byte)               "midi-clock")  
            ((midi-tick-event? status-byte)                "midi-tick" )  
            ((midi-start-event? status-byte)               "midi-start" ) 
            ((midi-continue-event? status-byte)            "midi-continue" )
            ((midi-stop-event? status-byte)                "midi-stop" )    
            ((midi-active-sense? status-byte)              "active-sense" ) 

            ((meta-event? status-byte)                      
               (string-append "Meta event " (as-string (string-length rest-bytes)))
               
            )
            (else (string-append "NON-EXISTING!" (as-string status-byte)))
          )
        )
  )
)  

; status-byte is a normal, decimal integer, between 0x80 and 0xFF (between 128 and 255).
; Return the least four significant bits, as a decimal number.
(define (channel-of-status-byte status-byte)
  (+ (remainder status-byte 16) 1)  ; Add 1, because the lowest channel is 1, not 0.
)

; Extract byte number n from str, which represents binary data (as a string of bytes).
; Returns an integer.
(define (get-byte-number str n)
  (as-number (string-ref str n)))

; Extract byte number n from str, which represents binary data (as a string of bytes).
; Returns a string.
(define (get-byte-number-as-string str n)
  (as-string (get-byte-number str n)))

(define (present-rest-bytes rest-bytes)
  (let ((lgt (string-length rest-bytes)))
    (cond ((= 1 lgt)  (as-string (as-number (string-ref rest-bytes 0))))
          ((= 2 lgt)  (string-append (get-byte-number-as-string rest-bytes 0) " " (get-byte-number-as-string rest-bytes 1) ))
          (else "..."))))
  
    
    
; ===============================================================================================================
; TRANSFORMATION OF INTERMEDIATE LIST STRUCTURE TO XML-IN-LAML AST.
; ...-as-list-to-ast functions.


; (define (midi-file-as-list-to-ast midi-file-list mode  non-midi-suffix-string)
;   (make-ast "StandardMidiFile"    ; use make-ast to avoid calling the mirror function StandardMidiFile. We do not wish to perfom the action procedure.
;     (cons 
;       (midi-header-chunck-as-list-to-ast (header-of-smf midi-file-list) mode)
;       (cons #t (map (lambda (track) (midi-track-chunck-as-list-to-ast track mode)) (tracks-of-smf midi-file-list)))
;     )
;     '()  ; attributes
;     'double
;     "midi"))

(define (midi-file-as-list-to-ast midi-file-list mode non-midi-suffix-string)
  (set! global-ppqn (ppqn-of-midi-header (header-of-smf midi-file-list)))
  (set! global-signature (find-time-signature-in-midi-list midi-file-list))
  (StandardMidiFile 'internal:run-action-procedure "false"
    (midi-header-chunck-as-list-to-ast (header-of-smf midi-file-list) mode)
    (map (lambda (track) (midi-track-chunck-as-list-to-ast track mode)) (tracks-of-smf midi-file-list))
    (if (empty-string? non-midi-suffix-string)
        '()
        (NonMidiSuffix (binary-to-hex-string non-midi-suffix-string)))))

; Attempt to find the time signature in midi-file-list (we use it in the info strings).
; Return a list, such as (4 4).
; If the time signature is not found, return (4 4) as default.
(define (find-time-signature-in-midi-list midi-file-list)
  (let ((tracks (tracks-of-smf midi-file-list))
        (default-value (list 4 4)))
    (if (not (null? tracks))
      (let* ((first-track (first tracks))
             (meta-time-sign-events (filter time-signature-event? (data-of-midi-track first-track))))
        (if (not (null? meta-time-sign-events))
            (extract-time-signature-of-track-event (first meta-time-sign-events))
            default-value))
      default-value)))

(define (time-signature-event? x)
  (if (and (list? x) (= (length x) 3) (eq? (first x) 'track-event))
      (let* ((me (midi-event-of-track-event x))
             (sb (status-byte-of-midi-event me))
             (dt (rest-bytes-of-midi-event me))
            )
        (if (= sb 255) ; meta event
            (let ((type (get-byte-number dt 0)))
               (if (= type 88) ; time signature
                   #t
                   #f))
            #f))
      #f))

(define (extract-time-signature-of-track-event track-event)
  (let* ((me (midi-event-of-track-event track-event))
         (dt (rest-bytes-of-midi-event me))
         (num (as-number (string-ref dt 2)))
         (denom (as-number (string-ref dt 3))))
    (list num (power 2 denom))))


; Convenient top-level function for transformation of intermediate midi list file structure to an AST.
(define midi-list-to-ast midi-file-as-list-to-ast)

(define laml-midi-file-prefix
  "(load (string-append laml-dir \"laml.scm\"))       
   (laml-style \"xml-in-laml/midi/midi\")
   (load (string-append laml-dir \"styles/xml-in-laml/midi/midi-laml-processing-lib.scm\"))"
)

; Given a midi ast, render it as XML-in-LAML at a file named output-file-path.
; output-file-path is the full path the output file, including extension.
(define (make-midi-laml-file midi-ast output-file-path)
  (if (file-exists? output-file-path) (delete-file output-file-path))
  (let* ((op (open-output-file output-file-path)))
     (render-as-laml-to-output-port midi-ast op laml-midi-file-prefix)
     (close-output-port op))
  (midi-report "Linearized LAML source string has been written to file. Now pretty printing...")
  (lib-load "file-read.scm")
  (lib-load "scheme-pretty-printing.scm") 
  (set! use-single-lining #t)
  (set! indentation-delta 3)
  (set! prefered-maximum-width 180)
  (pretty-print-lisp-file output-file-path output-file-path)
  (midi-report (string-append "The file " output-file-path " has been created.")  )
  (midi-report "You can edit the LAML file, process it, and reconstruct a midi file from it.")  
)

(define (midi-header-chunck-as-list-to-ast midi-header-chunck mode)
  (let ((format (format-of-midi-header midi-header-chunck))
        (number-of-tracks (number-of-tracks-of-midi-header midi-header-chunck))
        (ppqn (ppqn-of-midi-header midi-header-chunck)))
    (MidiHeader 'format (as-string format) 'numberOfTracks (as-string number-of-tracks)
                'pulsesPerQuarterNote (as-string ppqn) 'mode (as-string mode) 'counterTransposition (as-string counter-transposition))))


; Convert a single track (on list form) to a MidiTrack AST.
;; mode is either raw, absTime or deltaTime (a symbol).
;; raw: delta times with NoteOn and NoteOff.
;; absTime: absolute times with durations.
;; deltaTime: delta times with duration.
(define (midi-track-chunck-as-list-to-ast midi-track-chunck mode)
 (cond ((eq? mode 'raw)
          (MidiTrack
           (map midi-track-event-as-list-to-ast (data-of-midi-track midi-track-chunck))))
       ((eq? mode 'absTime)
          (track-abs-time-on-off-to-abs-time-with-duration
           (track-delta-to-abs-time
            (MidiTrack
             (map midi-track-event-as-list-to-ast (data-of-midi-track midi-track-chunck))))))
       ((eq? mode 'absTime-on-off)
         (track-delta-to-abs-time
            (MidiTrack
             (map midi-track-event-as-list-to-ast (data-of-midi-track midi-track-chunck)))))
       ((eq? mode 'deltaTime)
          (track-abs-time-with-duration-to-delta-time-with-duration
           (track-abs-time-on-off-to-abs-time-with-duration
            (track-delta-to-abs-time
             (MidiTrack
              (map midi-track-event-as-list-to-ast (data-of-midi-track midi-track-chunck))))) 0))
       (else (laml-error "midi-track-chunck-as-list-to-ast: Unknown mode:" mode))))
       

; Convert a single track (on list form) to an AST.
; Sample transformations back and forth:
; (define (midi-track-chunck-as-list-to-ast midi-track-chunck)
;    (track-abs-to-delta-time
;     (track-abs-time-with-duration-to-abs-time-on-off
;      (track-delta-with-duration-to-abs-time-with-duration
;       (track-abs-time-with-duration-to-delta-time-with-duration
;        (track-abs-time-on-off-to-abs-time-with-duration
;         (track-delta-to-abs-time
;          (MidiTrack
;           (map midi-track-event-as-list-to-ast (data-of-midi-track midi-track-chunck))))) 0 ))) 0))

; An amount of transposition introduced to show note info that corresponds to actual playing at the keyboard.
(define counter-transposition 0)


; Convert at single track-event, as constructed by make-track-event, to an AST.
; There are many different forms of the resulting AST!
(define (midi-track-event-as-list-to-ast track-event)
  (let* ((delta-time (delta-time-of-track-event track-event))    ; int
         (midi-event (midi-event-of-track-event track-event))    
         (status-byte (status-byte-of-midi-event midi-event))    ; int
         (rest-bytes (rest-bytes-of-midi-event midi-event))      ; binary data - byte string
        )
   (cond ((note-off-event? status-byte)                (NoteOff 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) 
                                                               'note (get-byte-number-as-string rest-bytes 0) 'velocity (get-byte-number-as-string rest-bytes 1) 
                                                               'info (note-number-to-note-name (as-number (get-byte-number-as-string rest-bytes 0)))
                                                       ))

         ((note-on-event? status-byte)                 (let ((note-number (get-byte-number rest-bytes 0))
                                                             (channel-number (channel-of-status-byte status-byte)))
                                                         (NoteOn 'deltaTime (as-string delta-time) 'channel (as-string channel-number) 
                                                                 'note (as-string note-number) 'velocity (get-byte-number-as-string rest-bytes 1)
                                                                 'info (string-append
                                                                        (note-number-to-note-name (as-number (get-byte-number-as-string rest-bytes 0)))
                                                                        (if (and info-about-drum-sounds? (drum-sound-channel? channel-number))
                                                                            (string-append " " (drum-info-string note-number))
                                                                            "")))
                                                       ))
         ((polyphonic-aftertouch-event? status-byte)   (PolyKeyPressure 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) 
                                                               'note (get-byte-number-as-string rest-bytes 0) 'pressure (get-byte-number-as-string rest-bytes 1))) 
         
         ((control-change-event? status-byte)          
             (let ((control-number (get-byte-number rest-bytes 0))
                   (control-value  (get-byte-number rest-bytes 1)))
               (cond ((= control-number 123)       (AllSoundOff 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) ))
                     ((= control-number 124)       (OmniOff 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) ))
                     ((= control-number 125)       (OmniOn 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) ))
                     ((= control-number 126)       (MonoMode 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) ))  ; ??
                     ((= control-number 127)       (PolyMode 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) ))  ; ??
                     (else                         (ControlChange 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte))
                                                                  'control (as-string control-number) 'value (as-string control-value) 
                                                                  'info (control-info-string control-number control-value)))
               )
             )              
         )
 
         ((program-change-event? status-byte)          (ProgramChange 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) 
                                                               'number (get-byte-number-as-string rest-bytes 0)) ) 
         ((channel-pressure-event? status-byte)        (ChannelKeyPressure 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) 
                                                               'pressure (get-byte-number-as-string rest-bytes 0))) 
         ((pitch-wheel-event? status-byte)             (PitchBendChange 'deltaTime (as-string delta-time) 'channel (as-string (channel-of-status-byte status-byte)) 
                                                               'value (as-string (+ (* (get-byte-number rest-bytes 0) 256) (get-byte-number rest-bytes 1))) ) )
         (else                        
          (cond
           ((system-exclusive-event? status-byte)           (SysEx 'deltaTime (as-string delta-time) (binary-to-hex-string rest-bytes)) )
                                                                                                           ; most probably: Take the last byte of rest-bytes away.

           ((mtc-quater-frame-message-event? status-byte)   (MTCQuarterFrame 'deltaTime (as-string delta-time) 'value (get-byte-number-as-string rest-bytes 0)) ) 
           ((song-position-pointer-event? status-byte)      (SongPositionPointer 'deltaTime (as-string delta-time) 'position ...) )
           ((song-select-event? status-byte)                (SongSelect 'deltaTime (as-string delta-time) 'number ...) )
           ((tune-request-event? status-byte)               (TuneRequest 'deltaTime (as-string delta-time)) ) 
           ((midi-clock-event? status-byte)                 (TimingClock 'deltaTime (as-string delta-time)) ) 
           ((midi-tick-event? status-byte)                  (laml-error "Encountered problematic midi-tick-event") )  ; !!??
           ((midi-start-event? status-byte)                 (Start  'deltaTime (as-string delta-time)) ) 
           ((midi-continue-event? status-byte)              (Continue 'deltaTime (as-string delta-time)) ) 
           ((midi-stop-event? status-byte)                  (Stop  'deltaTime (as-string delta-time))) 
           ((midi-active-sense? status-byte)                (ActiveSensing 'deltaTime (as-string delta-time)) ) 
           ((meta-event? status-byte)                       
                                                            (let* ((type (get-byte-number-as-string rest-bytes 0))
                                                                   (vlq-info (parse-variable-length-quantity rest-bytes 1))
                                                                   (vlq-length (car vlq-info))
                                                                   (vlq (cdr vlq-info))
                                                                   (rest-bytes-lgt (string-length rest-bytes))
                                                                   (rest-bytes (substring rest-bytes (+ 1 vlq-length) rest-bytes-lgt))
                                                                  )
                                                              (Meta 'deltaTime (as-string delta-time) 
                                                                    'type type
                                                                     'info (meta-info-string (as-number type) rest-bytes)
                                                                     (if (member (as-number type) textual-meta-types)
                                                                         rest-bytes
                                                                         (binary-to-hex-string 
                                                                          rest-bytes)))) )
           (else (laml-error "midi-track-event-as-list-to-ast: Non-existing system common, real-time and system exclusive event. Status-byte: " (as-string status-byte)))
          )
         )
   )
  )
)

(define (drum-info-string note)
  (case note
     (( 13)  "Latin Perc. Surdo Mute")
     (( 14)  "Latin Perc. Surdo open")
     (( 15)  "Hi Q")
     (( 16)  "Whip Slap ")
     (( 17)  "Scratch H")
     (( 18)  "Scratch L")
     (( 19)  "Finger Snap")
     (( 20)  "Click noice")
     (( 21)  "Metronome clock")
     (( 22)  "Metroneme bell")
     (( 23)  "Seq Click L")
     (( 24)  "Seq Click H")
     (( 25)  "Snare. Brush tab")
     (( 26)  "Snare. Brush swirl")
     (( 27)  "Snare. Brush slap")
     (( 28)  "Snare. Brush tap swirl")
     (( 29)  "Snare Rool")
     (( 30)  "Castanet")
     (( 31)  "Snare soft")
     (( 32)  "Sticks")
     (( 33)  "Bass drum. Kick soft")
     (( 34)  "Snare drum. Open rim shot")
     (( 35)  "Acoustic Bass Drum ")
     (( 36)  "Bass Drum 1")
     (( 37)  "Side Stick")
     (( 38)  "Acoustic Snare")
     (( 39)  "Hand Clap")
     (( 40)  "Electric Snare")
     (( 41)  "Low Floor Tom")
     (( 42)  "Closed Hi-Hat")
     (( 43)  "High Floor Tom")
     (( 44)  "Pedal Hi-Hat")
     (( 45)  "Low Tom")
     (( 46)  "Open Hi-Hat")
     (( 47)  "Low-Mid Tom")
     (( 48)  "Hi-Mid Tom")
     (( 49)  "Crash Cymbal 1")
     (( 50)  "High Tom")
     (( 51)  "Ride Cymbal 1")
     (( 52)  "Chinese Cymbal")
     (( 53)  "Ride Bell")
     (( 54)  "Tambourine")
     (( 55)  "Splash Cymbal")
     (( 56)  "Cowbell")
     (( 57)  "Crash Cymbal 2")
     (( 58)  "Vibraslap")
     (( 59)  "Ride Cymbal 2")
     (( 60)  "Hi Bongo")
     ((61)  "Low Bongo")
     ((62)  "Mute Hi Conga")
     ((63)  "Open Hi Conga")
     ((64)  "Low Conga")
     ((65)  "High Timbale")
     ((66)  "Low Timbale")
     ((67)  "High Agogo")
     ((68)  "Low Agogo")
     ((69)  "Cabasa")
     ((70)  "Maracas")
     ((71)  "Short Whistle")
     ((72)  "Long Whistle")
     ((73)  "Short Guiro")
     ((74)  "Long Guiro")
     ((75)  "Claves")
     ((76)  "Hi Wood Block")
     ((77)  "Low Wood Block")
     ((78)  "Mute Cuica")
     ((79)  "Open Cuica")
     ((80)  "Mute Triangle")
     ((81)  "Open Triangle")
     ((82)  "Shaker")
     ((83)  "Jingle Bells")
     ((84)  "Bell tree") 
     (else "")))
    

(define (meta-info-string type rest-bytes)
  (cond ((= type 0) "Sequence number")
        ((= type 1) "Text") 
        ((= type 2) "Copyright") 
        ((= type 3) "Sequence/track name") 
        ((= type 4) "Instrument name") 
        ((= type 5) "Lyric") 
        ((= type 6) "Marker")         
        ((= type 7) "Cue point") 
        ((= type 8) "Program name (patch/voice)")         
        ((= type 9) "Devise/port name") 
        ((= type 47) "End of track")         
        ((= type 81) (string-append "Tempo: " (as-string (byte-string-to-integer rest-bytes)) " microseconds per quarter note")) 
        ((= type 84) "SMPTE Offset: hh mm se fr ff")         
        ((= type 88) (string-append "Time signature " (time-signature-info rest-bytes)  ))         
        ((= type 89) (string-append "Key Signature " (key-signature-info rest-bytes)))

        ; Chord Name meta
        ((and (= type 127) (= (string-length rest-bytes) 7)
              (= 67 (get-byte-number rest-bytes 0))  ; 43 hex
              (= 123 (get-byte-number rest-bytes 1)) ; 7B hex
              (= 1 (get-byte-number rest-bytes 2)))   ; 01 hex     
           (chord-name-info (get-byte-number rest-bytes 3) (get-byte-number rest-bytes 4)))

        ((= type 127) "Proprietary event")         
        (else "???")         
  )
)

; cr is chord root (encoded) and cr is chord type number
(define (chord-name-info cr ct)
 (let ((chord-type-name (lookup-chord-type-short-name ct)))
  (string-append "Chord: " 
    (chord-root-info cr)
    (if (equal? chord-type-name "M") ; major is default, and it is not shown explicitly
        ""
        chord-type-name))))


(define (chord-root-info cr)
  (let ((cr-most-significant-half-byte    (quotient cr 16))
        (cr-least-significant-half-byte (remainder cr 16)))
    (string-append
      (cond ((= cr-least-significant-half-byte 0) "??")
            ((= cr-least-significant-half-byte 1) "C")
            ((= cr-least-significant-half-byte 2) "D")
            ((= cr-least-significant-half-byte 3) "E")
            ((= cr-least-significant-half-byte 4) "F")
            ((= cr-least-significant-half-byte 5) "G")
            ((= cr-least-significant-half-byte 6) "A")
            ((= cr-least-significant-half-byte 7) "B")
            (else "??"))
      (cond ((= cr-most-significant-half-byte 0) "bbb")
            ((= cr-most-significant-half-byte 1) "bb")
            ((= cr-most-significant-half-byte 2) "b")
            ((= cr-most-significant-half-byte 3) "")   ; natural
            ((= cr-most-significant-half-byte 4) "#")
            ((= cr-most-significant-half-byte 5) "##")
            ((= cr-most-significant-half-byte 6) "###")
            (else "??")))))





(define (key-signature-info rest-bytes)
  (let ((sf (as-number (string-ref rest-bytes 0)))
        (mi (as-number (string-ref rest-bytes 1)))
       ) 
   (string-append
       (as-string sf) " " 
       (cond ((= mi 1) "minor")
             ((= mi 0) "major")
             (else "???")))))

(define (time-signature-info str)
  (let* ((num (as-number (string-ref str 0)))
         (denom (as-number (string-ref str 1)))
         (number-of-midi-clock-in-metronome-click (as-number (string-ref str 2))) ; not shown
                                        ; one additional byte is present
        )
    (string-append 
     (as-string num) ":" (as-string (power 2 denom)))))

; Return n * n * ... n (m times)
(define (power n m)
  (if (= m 0) 1 (* n (power n (- m 1)))))

; Return short informal, intuitive description of a control change message with a given
; control-number and control-value.
(define (control-info-string control-number control-value)
  (cond ((= control-number 0) "Bank select MSB")
        ((= control-number 1)  "Modulation wheel MSB")
        ((= control-number 2)  "Breath control MSB")
        ((= control-number 3)  "Undefined")
        ((= control-number 4)  "Foot control MSB")
        ((= control-number 5)  "Portamento time MSB")
        ((= control-number 6)  "Data entry MSB")
        ((= control-number 7)  "Channel volume MSB")
        ((= control-number 8)  "Balance MSB")
        ((= control-number 9)  "Undefined")
        ((= control-number 10)  "Pan MSB")
        ((= control-number 11)  "Expression controller MSB")
        ((= control-number 12)  "Effect control 1 MSB")
        ((= control-number 13)  "Effect control 2 MSB")
        ((= control-number 14)  "Undefined")
        ((= control-number 15)  "Undefined")
        ((= control-number 16)  "General purpose controller 1 MSB")
        ((= control-number 17)  "General purpose controller 2 MSB")
        ((= control-number 18)  "General purpose controller 3 MSB")
        ((= control-number 19)  "General purpose controller 4 MSB")
        ((and (>= control-number 20) (<= control-number 31))  "Undefined")
        ((= control-number 32) "Bank select LSB")
        ((= control-number 33)  "Modulation wheel LSB")
        ((= control-number 34)  "Breath control LSB")
        ((= control-number 35)  "Undefined")
        ((= control-number 36)  "Foot control LSB")
        ((= control-number 37)  "Portamento time LSB")
        ((= control-number 38)  "Data entry LSB")
        ((= control-number 39)  "Channel volume LSB")
        ((= control-number 40)  "Balance LSB")
        ((= control-number 41)  "Undefined")
        ((= control-number 42)  "Pan LSB")
        ((= control-number 43)  "Expression controller LSB")
        ((= control-number 44)  "Effect control 1 LSB")
        ((= control-number 45)  "Effect control 2 LSB")
        ((= control-number 46)  "Undefined")
        ((= control-number 47)  "Undefined")
        ((= control-number 48)  "General purpose controller 1 LSB")
        ((= control-number 49)  "General purpose controller 2 LSB")
        ((= control-number 50)  "General purpose controller 3 LSB")
        ((= control-number 51)  "General purpose controller 4 LSB")
        ((and (>= control-number 52) (<= control-number 63))  "Undefined")
        ((= control-number 64)  (string-append "Hold pedal (sustain) " (on-off control-value)) )
        ((= control-number 65)  (string-append "Portamento " (on-off control-value)))
        ((= control-number 66)  (string-append "Sustenuto " (on-off control-value)))
        ((= control-number 67)  (string-append "Soft pedal " (on-off control-value))"")
        ((= control-number 69)  "Hold 2")
        ((= control-number 70)  "Sound controller 1: Sound Variation")
        ((= control-number 71)  "Sound controller 2: Timbre")
        ((= control-number 72)  "Sound controller 3: Release time")
        ((= control-number 73)  "Sound controller 4: Attack time")
        ((= control-number 74)  "Sound controller 5: Brightness")
        ((= control-number 75)  "Sound controller 6")
        ((= control-number 77)  "Sound controller 7")
        ((= control-number 78)  "Sound controller 8")
        ((= control-number 79)  "Sound controller 9")
        ((= control-number 80)  (string-append "General purpose button 5 " (on-off control-value)))
        ((= control-number 81)  (string-append "General purpose button 6 ") (on-off control-value))
        ((= control-number 82)  (string-append "General purpose button 7 ") (on-off control-value))
        ((= control-number 83)  (string-append "General purpose button 8 ") (on-off control-value))
        ((= control-number 84)  "Portamento Control")
        ((and (>= control-number 85) (<= control-number 90))  "Undefined")
        ((= control-number 91)  "Reverb send level")
        ((= control-number 92)  "Effects 2 depth")
        ((= control-number 93)  "Chorus send level ")
        ((= control-number 94)  "Effects 4 depth")
        ((= control-number 95)  "Effects 5 depth")
        ((= control-number 96)  "Data entry +1")
        ((= control-number 97)  "Data entry -1")
        ((= control-number 98)  "Non-registered parameter number LSB")
        ((= control-number 99)  "Non-registered parameter number MSB")
        ((= control-number 100)  "Registered parameter number LSB")
        ((= control-number 101)  "Registered parameter number MSB")
        ((and (>= control-number 102) (<= control-number 119))  "Undefined")
        ; The rest have special ASTs.
        (else "???")
   )
)

(define (on-off control-value)
  (if (< control-value 63) "off" "on"))

; The meta types for which we do NOT convert the contents to a hex string.
(define textual-meta-types
  (list 1 2 3 4 5 6 7 8 9))



; ===============================================================================================================
; TRANSFORMATION OF XML-IN-LAML ASTs to binary MIDI format. 
; Not done via the intermediate list format. Functions return strings of bytes (binary information).
; ...-ast-to-bin functions.

; Top level function that transforms an overall standard midi file AST to the binary midi format.
; Returns a string (of bytes) - binary data. 
(define (standard-midi-file-ast-to-bin smf-ast)
  (let* (
         (header-ast (ast-subtree smf-ast "MidiHeader"))
         (format (as-number (ast-attribute header-ast 'format)))
         (target-format (as-number (ast-attribute header-ast 'target-format format)))

         (target-ppqn (as-number (ast-attribute header-ast 'target-pulsesPerQuarterNote (ast-attribute header-ast 'pulsesPerQuarterNote))))
         (ppqn (as-number (ast-attribute header-ast 'pulsesPerQuarterNote)))
         (ppqn-factor (/ target-ppqn ppqn))

         (smf-ast-1 (cond 
                      ((= format target-format) smf-ast)
                      ((and (= format 1) (= target-format 0)) (track-joined-standard-midi-file-ast smf-ast))
                      (else (laml-error "standard-midi-file-ast-to-bin: Illegal target-format relative to format:" target-format format))))

         (header-ast-1 (ast-subtree smf-ast-1 "MidiHeader"))
         (header-ast-2 (copy-ast-mutate-attributes header-ast-1 'pulsesPerQuarterNote (as-string target-ppqn)))
         (mode-symbol (as-symbol (ast-attribute header-ast-2 'mode)))
                           
         (counter-transp (as-number (ast-attribute header-ast-2 'counterTransposition "0")))
         (track-asts (filter (lambda (x) (and (ast? x) (equal? (ast-element-name x) "MidiTrack"))) (ast-subtrees smf-ast-1)))  ; a list of MidiTrack asts
         (transformed-track-asts (map (lambda (track-ast) (transform-track-to-raw-mode track-ast mode-symbol ppqn-factor)) track-asts))  ; raw mode ast
 
         (non-midi-suffix-ast (ast-subtree smf-ast "NonMidiSuffix"))
         (non-midi-suffix-hex-string (if non-midi-suffix-ast 
                                     (ast-text non-midi-suffix-ast)
                                     ""))
        )
    (set! counter-transposition counter-transp)
    (string-append
      (midi-header-ast-to-bin header-ast-2)
      (midi-tracks-ast-to-bin transformed-track-asts)   ; transform raw ast to binary form
      (hex-to-binary-string-relaxed non-midi-suffix-hex-string)
    )))        

; Transform a track ast to raw mode from whatever mode mode-system prescribes.
(define (transform-track-to-raw-mode midi-track-ast mode-symbol ppqn-factor)
  (cond ((eq? mode-symbol 'raw)
            midi-track-ast)                                                             ; the identity transformation
        ((eq? mode-symbol 'absTime)
            (track-abs-to-delta-time
              (track-abs-time-with-duration-to-abs-time-on-off midi-track-ast) 0 ppqn-factor ))
        ((eq? mode-symbol 'absTime-on-off)
            (track-abs-to-delta-time
              midi-track-ast 0 ppqn-factor ))
        ((eq? mode-symbol 'deltaTime)
            (track-abs-to-delta-time
              (track-abs-time-with-duration-to-abs-time-on-off
                (track-delta-with-duration-to-abs-time-with-duration midi-track-ast)) 0 ppqn-factor))
        (else (laml-error "transform-track-to-raw-mode: Unknown mode" mode-symbol))))

(define (midi-header-ast-to-bin header-ast)
  (let ((format (ast-attribute header-ast 'format))
        (number-of-tracks (ast-attribute header-ast 'numberOfTracks))
        (ppqn (ast-attribute header-ast 'pulsesPerQuarterNote))
       )
    (string-append "MThd" 
                   (int10-to-binary 6 4)                ; Length of remaining: always 6
                   (int10-to-binary (as-number format) 2)           ; First 2 bytes of the 6 bytes
                   (int10-to-binary (as-number number-of-tracks) 2) ; Next 2 bytes of the 6 bytes
                   (int10-to-binary (as-number ppqn) 2)             ; Final 2 bytes of the 6 bytes
    )))

(define (midi-tracks-ast-to-bin track-ast-list)  ; plural - a list of MidiTrack asts
  (let* ((track-string-list (map midi-one-track-ast-to-bin track-ast-list)))
     (apply string-append track-string-list)))

(define (midi-one-track-ast-to-bin track-ast)   ; track-ast is a MidiTrack ast.
  (let* ((messages-asts (filter ast? (ast-subtrees track-ast)))
         (track-string-list (map single-midi-track-event-ast-to-bin (filter legal-channel? messages-asts)))
         (track-string (apply string-append track-string-list))
        )
    (string-append "MTrk"
                   (int10-to-binary (string-length track-string) 4) ; 4 bytes length of data
                   track-string)))

; mes-ast is legal if it is a non-channel message, or a channel message assigned to a channel between 1 and 16.
(define (legal-channel? mes-ast)
  (let ((ch (ast-attribute mes-ast 'channel #f)))
    (if ch  ; a channel message
        (let ((ch-number (as-number ch)))
           (and (>= ch-number 1) (<= ch-number 16)))
        #t)))
               



; Convert a single midi track event ast (such as a NoteOn ast) to binary form.
; Return a string (binary data).
(define (single-midi-track-event-ast-to-bin ast)
 (letrec ((ann (lambda (ast name) (as-number (ast-attribute ast name)))))  ; attribute name as number
  (string-append
   (unparse-integer-as-variable-length-quantity (ann ast 'deltaTime))                  ; deltaTime first, as variable length quantity
   (cond  ((equal? (ast-element-name ast) "NoteOff") 
            (string-append 
               (make-byte-string-from-hex-2 8 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'note) 1)                                     ; one byte, note
               (int10-to-binary (ann ast 'velocity) 1)                                 ; one byte, velocity
            ))
         ((equal? (ast-element-name ast) "NoteOn")  
            (string-append 
               (make-byte-string-from-hex-2 9 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'note) 1)                                     ; one byte, note
               (int10-to-binary (ann ast 'velocity) 1)                                 ; one byte, velocity
            )
         ) 
         ((equal? (ast-element-name ast) "PolyKeyPressure") 
            (string-append 
               (make-byte-string-from-hex-2 10 (- (ann ast 'channel) 1))               ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'note) 1)                                     ; one byte, note
               (int10-to-binary (ann ast 'pressure) 1)                                 ; one byte
            )
         ) 
         ((equal? (ast-element-name ast) "ControlChange")    
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))               ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'control) 1)                                  ; one byte
               (int10-to-binary (ann ast 'value) 1)                                    ; one byte
            )
         ) 
         ((equal? (ast-element-name ast) "ProgramChange") 
            (string-append 
               (make-byte-string-from-hex-2 12 (- (ann ast 'channel) 1))               ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'number) 1)                                   ; one byte
            )
         ) 
         ((equal? (ast-element-name ast) "ChannelKeyPressure") 
            (string-append 
               (make-byte-string-from-hex-2 13 (- (ann ast 'channel) 1))               ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'pressure) 1)                                 ; one byte
            )
         ) 
         ((equal? (ast-element-name ast) "PitchBendChange")
            (string-append 
               (make-byte-string-from-hex-2 14 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary (ann ast 'value) 2)                                     ; the value attribute is large integer, represented by two bytes!
                                                                                        ; Is the two bytes correctly constructed???
            )
         )
         ((equal? (ast-element-name ast) "AllSoundOff")    
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 120 1)                                                  ; one byte. Fixed control number
               (int10-to-binary 0 1)                                                    ; one byte. Not used, defaults to 0
            )
         )
         ((equal? (ast-element-name ast) "ResetAllControllers") 
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 121 1)                                                  ; one byte. Fixed control number
               (int10-to-binary 0 1)                                                    ; one byte. Not used, defaults to 0
            )
         )
         ((equal? (ast-element-name ast) "LocalControl")  
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 122 1)                                                  ; one byte. Fixed control number
               (int10-to-binary (if (equal? (ast-attribute ast 'value) "off") 0 127))   ; one byte. 
            )
         )
         ((equal? (ast-element-name ast) "AllNotesOff")
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 123 1)                                                  ; one byte. Fixed control number
               (int10-to-binary 0 1)                                                    ; one byte. Not used, defaults to 0
            )
         )
         ((equal? (ast-element-name ast) "OmniOff")   
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 124 1)                                                  ; one byte. Fixed control number
               (int10-to-binary 0 1)                                                    ; one byte. Not used, defaults to 0
            )
         )
         ((equal? (ast-element-name ast) "OmniOn") 
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 125 1)                                                  ; one byte. Fixed control number
               (int10-to-binary 0 1)                                                    ; one byte. Not used, defaults to 0
            )
         )
         ((equal? (ast-element-name ast) "MonoMode")  
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 126 1)                                                  ; one byte. Fixed control number
               (int10-to-binary (ann ast 'value) 1)                                     ; one byte.
            )
         )
         ((equal? (ast-element-name ast) "PolyMode") 
            (string-append 
               (make-byte-string-from-hex-2 11 (- (ann ast 'channel) 1))                ; status byte: status half byte + channel number
               (int10-to-binary 127 1)                                                  ; one byte. Fixed control number
               (int10-to-binary 0 1)                                                    ; one byte. Not used, defaults to 0
            )
         )
         ((equal? (ast-element-name ast) "ControlChange14") 
            (laml-error "Cannot emit ControlChange14 AST to binary midi string.") 
         ) 
         ((equal? (ast-element-name ast) "RPNChange")
            (laml-error "Cannot emit RPNChange AST to binary midi string.") 
         )
         ((equal? (ast-element-name ast) "SysEx")    
            (string-append
               (make-byte-string-from-hex-2 15 0)                                       ; one byte  F0. Start sys ex.
               (hex-to-binary-string (ast-text ast))                                    ; includes trailing F7.  Thus next line is not necssary. Therefore comment.
               ; (make-byte-string-from-hex-2 15 7)                                                   ; one byte  F7. End sys ex.
            )
         ) 
         ((equal? (ast-element-name ast) "MTCQuarterFrame")
            (string-append
               (make-byte-string-from-hex-2 15 1)                                       ; one byte  F1.
               (int10-to-binary (ann ast 'value) 1)                                     ; one data byte.               
            )
         )
         ((equal? (ast-element-name ast) "SongPositionPointer") 
            (string-append
               (make-byte-string-from-hex-2 15 2)                                       ; one byte  F2.
               (int10-to-binary (ann ast 'position) 2)                                  ; the value attribute is large integer, represented by two bytes!
                                                                                        ; Is the two bytes correctly constructed???
            )
         )
         ((equal? (ast-element-name ast) "SongSelect")   
            (string-append
               (make-byte-string-from-hex-2 15 3)                                       ; one byte  F3.
               (int10-to-binary (ann ast 'number) 1)                                    ; one data byte.               
            )
         )
         ((equal? (ast-element-name ast) "TuneRequest")  
            (string-append
               (make-byte-string-from-hex-2 15 6)                                       ; one byte  F6.
            )

         )
         ((equal? (ast-element-name ast) "TimingClock") 
            (string-append
               (make-byte-string-from-hex-2 15 8)                                       ; one byte  F8.
            )
         )
         ((equal? (ast-element-name ast) "Start")  
            (string-append
               (make-byte-string-from-hex-2 15 10)                                      ; one byte  FA.
            )
         )
         ((equal? (ast-element-name ast) "Continue") 
            (string-append
               (make-byte-string-from-hex-2 15 11)                                      ; one byte  FB.
            )
         )
         ((equal? (ast-element-name ast) "Stop") 
            (string-append
               (make-byte-string-from-hex-2 15 12)                                      ; one byte  FC.
            )
         )
         ((equal? (ast-element-name ast) "ActiveSensing")
            (string-append
               (make-byte-string-from-hex-2 15 14)                                      ; one byte  FE.
            )
         )         
         ((equal? (ast-element-name ast) "Meta") 
           (let* ((type-number (ann ast 'type))
                  (contents (ast-text ast))
                  (meta-bytes (if (member type-number textual-meta-types) contents (hex-to-binary-string contents)))   ; either ascii string or hex presented string 
                  (vlq-length (unparse-integer-as-variable-length-quantity (string-length meta-bytes)))
                 )
                  
            (string-append 
               (make-byte-string-from-hex-2 15 15)                                      ; one byte  FF.
               (int10-to-binary (ann ast 'type) 1)                                      ; one data byte. The type of the meta event
               vlq-length                                                               ; length of meta-bytes, as a variable length quantity
               meta-bytes                                                               ; the meta bytes as such
            )
           )
         )
         (else (laml-error "single-midi-track-event-ast-to-bin: Unknown element name of AST:" (ast-element-name ast)))

   ))))


; ===============================================================================================================
; Testing stuff.

; Compare two binary strings.
(define (binary-compare str1 str2)
 (let ((lgt1 (string-length str1))
       (lgt2 (string-length str2)))
  (if (not (= lgt1 lgt2)) (display-message "Not same length" lgt1 lgt2))
  (binary-compare-1 str1 str2 0 (string-length str1) (string-length str2))))
  

(define (binary-compare-1 str1 str2 i lgt1 lgt2)
  (if (< i (min lgt1 lgt2))
    (if (eqv? (string-ref str1 i) (string-ref str2 i))
        (binary-compare-1 str1 str2 (+ i 1) lgt1 lgt2)
        (begin (display-message "Difference at index" i (binary-to-hex-string (as-string (string-ref str1 i))) (binary-to-hex-string (as-string (string-ref str2 i))))
               (binary-compare-1 str1 str2 (+ i 1) lgt1 lgt2)))))



; ===============================================================================================================
; AST Transformations, related to deltaTime, absTime og duration.
; #1

; Convert deltaTime to absTime attributes in the MidiTrack ast midi-track-ast. 
; Removes deltaTime attributes and adds absTime attributes.
; Does not add duration.
; Returns a revised copy of midi-track-ast (a MidiTrack ast).
; Inverse function: track-abs-to-delta-time.
(define (track-delta-to-abs-time midi-track-ast)
  (let ((message-asts (filter ast? (ast-subtrees midi-track-ast))))
    (MidiTrack 
      (message-list-delta-time-to-abs-time message-asts 0))))

(define (message-list-delta-time-to-abs-time message-ast-list start-time)
  (cond ((null? message-ast-list) '())
        (else (let* ((message-ast (car message-ast-list))
                     (delta-time (ast-attribute message-ast 'deltaTime))
                     (new-abs-time (+ start-time (as-number delta-time)))
                    )
                (cons (single-message-ast-delta-to-abs-time message-ast new-abs-time)
                      (message-list-delta-time-to-abs-time (cdr message-ast-list) new-abs-time))))))

; A mutating alternative to the copying function next to this function. OK, but does not really affect the total running time
; (define (single-message-ast-delta-to-abs-time ast abs-time)
;  (let ((existing-info (ast-attribute ast 'info "")))
;   (set-ast-attributes! ast
;      (append (remove-props! '(deltaTime 'info) (ast-attributes ast))
;              (list
;               'absTime (as-string abs-time) 
;               'info (string-append (bar-beat-clock-string abs-time global-ppqn (first global-signature) (second global-signature)) " " existing-info)
;               )))
;   ast))

(define (single-message-ast-delta-to-abs-time ast abs-time)
 (let ((existing-info (ast-attribute ast 'info "")))
  (make-ast (ast-element-name ast) 
            (ast-subtrees ast) 
            (append (list
                       'absTime (as-string abs-time) 
                       'info (string-append (bar-beat-clock-string abs-time global-ppqn (first global-signature) (second global-signature)) " " existing-info)
                    )
                    (but-props (ast-attributes ast) (list 'deltaTime 'info)))
            (ast-kind ast)
            (ast-language ast)
            (ast-internal-attributes ast))))


; ----------------------------------------------------------------------------------------

; #6
; Convert absTime to deltaTime attributes in the MidiTrack ast midi-track-ast.
; Returns a revised copy of midi-track-ast (a MidiTrack ast).
; Inverse function: track-delta-to-abs-time.
; The parameter initial-delta-time is typically 0.
(define (track-abs-to-delta-time midi-track-ast initial-delta-time ppqn-factor)
  (let ((message-asts (filter ast? (ast-subtrees midi-track-ast))))
    (cond ((null? message-asts)
             (MidiTrack))
          (else 
             (MidiTrack
               (let* ((first-message-ast (first message-asts))
                      (first-abs-time (as-number (ast-attribute first-message-ast 'absTime))))
                 (cons (single-message-ast-abs-to-delta-time first-message-ast initial-delta-time ppqn-factor)
                       (message-list-abs-to-delta-time (cdr message-asts) first-abs-time ppqn-factor))))))))
      

; previous-abs-time is the absolute time of the previous message ast.
(define (message-list-abs-to-delta-time message-ast-list previous-abs-time ppqn-factor)
  (cond ((null? message-ast-list) '())
        (else (let* ((message-ast (car message-ast-list))
                     (this-abs-time (as-number (ast-attribute message-ast 'absTime)))
                     (new-delta-time (- this-abs-time previous-abs-time))
                    )
                (cons (single-message-ast-abs-to-delta-time message-ast new-delta-time ppqn-factor)
                      (message-list-abs-to-delta-time (cdr message-ast-list) this-abs-time ppqn-factor))))))

(define (single-message-ast-abs-to-delta-time ast delta-time ppqn-factor)
  (make-ast (ast-element-name ast) 
            (ast-subtrees ast) 
            (append (list 'deltaTime 
                           (if (= ppqn-factor 1)
                               (as-string delta-time)
                               (as-int-string (to-int (* ppqn-factor delta-time)))))
                    (but-props (ast-attributes ast) (list 'absTime)))
            (ast-kind ast)
            (ast-language ast)
            (ast-internal-attributes ast)))

; ----------------------------------------------------------------------------------------

; #2
; Convert absTime with NoteOn and NoteOff messages to absTime with duration.
; Removes all NoteOff messages.
; Returns a revised copy of midi-track-ast (a MidiTrack ast).
(define (track-abs-time-on-off-to-abs-time-with-duration midi-track-ast)
  (let ((message-asts (filter ast? (ast-subtrees midi-track-ast))))
    (MidiTrack 
      (messages-abs-time-on-off-to-abs-time-with-duration message-asts))))

(define (messages-abs-time-on-off-to-abs-time-with-duration message-ast-list)
  (cond ((null? message-ast-list) '())
        ((note-on-ast? (first message-ast-list)) 
           (let* ((current-ast (first message-ast-list))
                  (current-channel (ast-attribute current-ast 'channel))
                  (current-note (ast-attribute current-ast 'note))
                  (note-off-ast (find-in-list 
                                  (lambda (ast ) (and ast
                                                      (note-off-ast? ast) 
                                                      (equal? (ast-attribute ast 'channel) current-channel)
                                                      (equal? (ast-attribute ast 'note) current-note)
                                                      (not (equal? "true" (ast-internal-attribute ast 'seen "false")))))
                                 message-ast-list))
                  (duration (if note-off-ast 
                                (- (as-number (ast-attribute note-off-ast 'absTime)) (as-number (ast-attribute current-ast 'absTime)))
                                1000000))  ; should not happen
                 )
              (if note-off-ast (set-internal-ast-attribute! note-off-ast 'seen "true"))
              (if (not note-off-ast)  
                  (display-warning "NoteOn without corresponding NoteOff"))
              (cons (single-message-ast-abs-time-add-duration (first message-ast-list) duration)
                    (messages-abs-time-on-off-to-abs-time-with-duration (cdr message-ast-list)))))
 
        ((note-off-ast? (first message-ast-list))   ; drop it. Has alreadby been taken into account.
                    (messages-abs-time-on-off-to-abs-time-with-duration (cdr message-ast-list)))
        (else (cons (first message-ast-list)
                    (messages-abs-time-on-off-to-abs-time-with-duration (cdr message-ast-list))))))

(define (note-on-ast? x)
 (and (ast? x) (equal? "NoteOn" (ast-element-name x)) (not (equal? (ast-attribute x 'velocity) "0"))))

(define (note-off-ast? x)
 (or
   (and (ast? x) (equal? "NoteOff" (ast-element-name x)))
   (and (ast? x) (equal? "NoteOn" (ast-element-name x)) (equal? (ast-attribute x 'velocity) "0"))))

(define (meta-chord-ast? x)
  (if (and (ast? x) (equal? "Meta" (ast-element-name x)) (= 127 (as-number (ast-attribute x 'type -1))))
      (let* ((meta-contents (ast-text x))
             (binary-meta-contents (hex-to-binary-string-relaxed meta-contents)))
         (and (= (string-length binary-meta-contents) 7)
              (= 67 (get-byte-number binary-meta-contents 0))  ; 43 hex
              (= 123 (get-byte-number binary-meta-contents 1)) ; 7B hex
              (= 1 (get-byte-number binary-meta-contents 2))) )
      #f))

; Given that meta-ast is a meta chord ast (satisfying the predicate meta-chord-ast?) 
; return the cons of root and chord-type, as a cons pair of strings. 
; Example ("C" . "7")
(define (meta-chord-root-and-chordtype meta-chord-ast)
 (let* ((meta-contents (ast-text meta-chord-ast))
        (binary-meta-contents (hex-to-binary-string-relaxed meta-contents))
        (root-number (get-byte-number binary-meta-contents 3))
        (chord-type-number (get-byte-number binary-meta-contents 4)))
   (cons (chord-root-info root-number)                             ; maybe a b chord. We only support # chords.
         (lookup-chord-type-short-name chord-type-number))))

(define (single-message-ast-abs-time-add-duration ast duration)
  (make-ast (ast-element-name ast) 
            (ast-subtrees ast) 
            (append (ast-attributes ast) (list 'duration (as-string duration)))
            (ast-kind ast)
            (ast-language ast)
            (ast-internal-attributes ast)))

; ----------------------------------------------------------------------------------------

; #5
; Convert absTime with duration to absTime with NoteOn and NoteOff attributes.
; In between NoteOn with absTime it is allowed to have NoteOn with deltaTime.
; A NoteOn with deltaTime is relative to either the previous NoteOn with deltaTime, or the previous NoteOn with absTime.
; Re-introduces NoteOff messages.
; Makes sure that the resulting messages are properly sorted.
(define (track-abs-time-with-duration-to-abs-time-on-off midi-track-ast)
  (let ((message-asts (filter ast? (ast-subtrees midi-track-ast))))
    (MidiTrack 
      (stable-sort-list    ; sort-list happens not to be stable. In fact it is systematically unstable. Therefore we use stable-sort-list.
        (messages-abs-time-with-duration-to-abs-time-on-off (get-rid-of-delta-times message-asts 0))
        (lambda (ast1 ast2)
          (<= (as-number (ast-attribute ast1 'absTime)) (as-number (ast-attribute ast2 'absTime))))))))


; Eliminate deltaTime, keep absTime.
; (define (get-rid-of-delta-times message-ast-list previous-abs-time)
;   (cond ((null? message-ast-list) '())
;         ((and (note-on-ast? (first message-ast-list)) (ast-attribute (first message-ast-list)  'absTime))       ; absTime NoteOn
;           (let ((abs-time (as-number (ast-attribute (first message-ast-list) 'absTime))))
;            (cons (first message-ast-list) 
;                  (get-rid-of-delta-times (cdr message-ast-list) abs-time))))
;         ((and (note-on-ast? (first message-ast-list)) (ast-attribute (first message-ast-list) 'deltaTime))      ; deltaTime  NoteOn
;           (let ((delta-time (as-number (ast-attribute (first message-ast-list) 'deltaTime))))
;            (cons (single-message-ast-delta-to-abs-time (first message-ast-list) (+ delta-time previous-abs-time)) 
;                  (get-rid-of-delta-times (cdr message-ast-list) (+ delta-time previous-abs-time)))))
;         (else (let ((abs-time (as-number (ast-attribute (first message-ast-list) 'absTime))))
;                       (cons (first message-ast-list)
;                             (get-rid-of-delta-times (cdr message-ast-list) abs-time)) ))))

; Eliminate deltaTime, keep absTime.
(define (get-rid-of-delta-times message-ast-list previous-abs-time)
  (cond ((null? message-ast-list) '())
        ((ast-attribute (first message-ast-list) 'deltaTime)      ; any event with deltaTime
          (let ((delta-time (as-number (ast-attribute (first message-ast-list) 'deltaTime))))
           (cons (single-message-ast-delta-to-abs-time (first message-ast-list) (+ delta-time previous-abs-time) ) 
                 (get-rid-of-delta-times (cdr message-ast-list) (+ delta-time previous-abs-time)))))
        (else (let ((abs-time (as-number (ast-attribute (first message-ast-list) 'absTime))))
                      (cons (first message-ast-list)
                            (get-rid-of-delta-times (cdr message-ast-list) abs-time)) ))))
                    
  

(define (messages-abs-time-with-duration-to-abs-time-on-off message-ast-list)
  (cond ((null? message-ast-list) '())
        ((note-on-ast? (first message-ast-list)) 
           (let* ((current-ast (first message-ast-list))
                  (current-channel (ast-attribute current-ast 'channel))
                  (current-note (ast-attribute current-ast 'note))
                  (current-abstime (as-number (ast-attribute current-ast 'absTime)))
                  (current-duration (as-number (ast-attribute current-ast 'duration)))
                 )
            (cons 
               (single-message-ast-abs-time-remove-duration current-ast)     ; NoteOn without duration
               (cons
                 (NoteOn 'absTime (as-string (+ current-abstime current-duration)) 'channel current-channel
                         'note current-note 'velocity "0") ; NoteOff = NoteOn with zero velocity
                   (messages-abs-time-with-duration-to-abs-time-on-off (cdr message-ast-list))))))
        (else (cons (first message-ast-list)
                    (messages-abs-time-with-duration-to-abs-time-on-off (cdr message-ast-list))))))
    

(define (single-message-ast-abs-time-remove-duration ast)
  (make-ast (ast-element-name ast) 
            (ast-subtrees ast) 
            (but-props (ast-attributes ast) (list 'duration))
            (ast-kind ast)
            (ast-language ast)
            (ast-internal-attributes ast)))

; ----------------------------------------------------------------------------------------

; # 3
; Convert a track, absTime with duration to deltaTime with duration.
; Returns a revised copy of midi-track-ast (a MidiTrack ast).
; The parameter initial-delta-time is typically 0.
; Very similar to message-list-abs-to-delta-time.
(define (track-abs-time-with-duration-to-delta-time-with-duration midi-track-ast initial-delta-time)
  (let ((message-asts (filter ast? (ast-subtrees midi-track-ast))))
    (cond ((null? message-asts)
             (MidiTrack))
          (else 
             (MidiTrack
               (let* ((first-message-ast (first message-asts))
                      (first-abs-time (as-number (ast-attribute first-message-ast 'absTime))))
                 (cons (single-message-ast-abs-to-delta-time first-message-ast initial-delta-time 1)
                       (message-list-abs-with-duration-to-delta-with-duration (cdr message-asts) first-abs-time))))))))
      

; previous-abs-time is the absolute time of the previous message ast.
(define (message-list-abs-with-duration-to-delta-with-duration message-ast-list previous-abs-time)
  (cond ((null? message-ast-list) '())
        (else (let* ((message-ast (car message-ast-list))
                     (this-abs-time (as-number (ast-attribute message-ast 'absTime)))
                     (new-delta-time (- this-abs-time previous-abs-time))
                    )
                (cons (single-message-ast-abs-to-delta-time message-ast new-delta-time 1)
                      (message-list-abs-with-duration-to-delta-with-duration (cdr message-ast-list) this-abs-time))))))

; single-message-ast-abs-to-delta-time defined earlier.

; ----------------------------------------------------------------------------------------

; #4 
; Convert deltaTime with duration to absTime wit duration attributes in the MidiTrack ast midi-track-ast. 
; Removes deltaTime attributes and adds absTime attributes.
; Does not add duration.
; Returns a revised copy of midi-track-ast (a MidiTrack ast).
; Inverse function: track-abs-to-delta-time.
; Very similar to track-delta-to-abs-time.
(define (track-delta-with-duration-to-abs-time-with-duration midi-track-ast)
  (let ((message-asts (filter ast? (ast-subtrees midi-track-ast))))
    (MidiTrack 
      (message-list-delta-time-with-duration-to-abs-time-with-duration message-asts 0))))

(define (message-list-delta-time-with-duration-to-abs-time-with-duration message-ast-list start-time)
  (cond ((null? message-ast-list) '())
        (else (let* ((message-ast (car message-ast-list))
                     (delta-time (ast-attribute message-ast 'deltaTime))
                     (new-abs-time (+ start-time (as-number delta-time)))
                    )
                (cons (single-message-ast-delta-to-abs-time message-ast new-abs-time)
                      (message-list-delta-time-with-duration-to-abs-time-with-duration (cdr message-ast-list) new-abs-time))))))

; single-message-ast-delta-to-abs-time define before

; ===============================================================================================================
; Note names and note numbers

(define (note-number-to-note-name n0)
 (let ((n (+ n0 counter-transposition)))
  (let ((root-number (remainder n 12))
        (octave-number (- (quotient n 12) 2))  ; yamaha convention
       )
   (string-append (name-of-note root-number) (as-string octave-number) 
        (if (not (= counter-transposition 0))   ; display counter transposition
            (if (> (- counter-transposition) 0)
                (string-append "+" (as-string (- counter-transposition)))
                (as-string (- counter-transposition)))
            "")))))

(define (note-name-to-note-number nm)
  (let* ((lgt (string-length nm))
         (pos (find-in-string-by-predicate (lambda (ch) (memv (as-number ch) (list 48 49 50 51 52 53 54 55 56 57  45))) nm))   ; pos of '-' or ciffer
         (root-name (upcase-string (substring nm 0 pos)))
         (octave-number (as-number (substring nm pos lgt))))
     (+ (* (+ octave-number 2) 12) (number-of-note-name (upcase-string root-name)))))

(define (name-of-note n)
  (cond ((= n 0) "C")
        ((= n 1) "C#")
        ((= n 2) "D")
        ((= n 3) "D#")
        ((= n 4) "E")
        ((= n 5) "F")
        ((= n 6) "F#")
        ((= n 7) "G")
        ((= n 8) "G#")
        ((= n 9) "A")
        ((= n 10) "A#")
        ((= n 11) "B")
        (else (laml-error "name-of-note: Invalid number:" n))))

(define (number-of-note-name nm)
  (cond ((equal? nm "C") 0)
        ((equal? nm "C#") 1)
        ((equal? nm "D") 2)
        ((equal? nm "D#") 3)
        ((equal? nm "E") 4)
        ((equal? nm "F") 5)
        ((equal? nm "F#") 6)
        ((equal? nm "G") 7)
        ((equal? nm "G#") 8)
        ((equal? nm "A") 9)
        ((equal? nm "A#") 10)
        ((or (equal? nm "B") (equal? nm "H"))  11)
        (else (laml-error "number-of-note-name: Invalid name:" nm))))

;; A function reserved for used in XML-in-LAML midi source files.
;; In reality an alias of note-name-to-note-number, which compensates for counter transposition.
;; It means that you can use (nn "note-name"), where note-name is as shown in the info attribute.
(define (nn nm)
  (- (note-name-to-note-number nm) counter-transposition))
 

; ===============================================================================================================
; Bar, beat and clock

; Return a list of three integers: bar, beat and clock.
; ppqn is the pulses per quarter note.
; The signare of the song if n:m
(define (bar-beat-clock abs-time ppqn n m)
 (let* ((quarter-notes-per-bar (quotient (* n 4) m))
        (pp-per-bar (* quarter-notes-per-bar ppqn))
        (pp-per-beat (quotient pp-per-bar m))

        (bar (quotient abs-time pp-per-bar))
        (rest1 (remainder abs-time pp-per-bar))
        (beat (quotient rest1 pp-per-beat))
        (rest2 (remainder rest1 pp-per-beat)))
    (list bar beat rest2)))

(define (bar-beat-clock-string abs-time ppqn n m)
  (let ((bbc (bar-beat-clock abs-time ppqn n m)))
    (string-append (as-string (first bbc)) ":" (as-string (second bbc)) ":" (as-string (third bbc))))) 


; ---------------------------------------------------------------------------------------------------------------
; Tyros Multipad stuff.
; It is assumed that we deal with multipad stuff in absTime mode.

(define (multipad-ast-to-standard-midi-file-ast ast)
  (let* ((mode (ast-attribute ast 'mode))
         (default-image-name "S809")

         (name-1    (ast-attribute ast 'name-1))
         (repeat-1  (as-boolean (ast-attribute ast 'repeat-1 "false")))
         (chord-match-1 (as-boolean (ast-attribute ast 'chord-match-1 "false")))
         (image-1 (ast-attribute ast 'image-1 default-image-name))

         (name-2 (ast-attribute ast 'name-2))
         (repeat-2  (as-boolean (ast-attribute ast 'repeat-2 "false")))
         (chord-match-2 (as-boolean (ast-attribute ast 'chord-match-2 "false")))
         (image-2 (ast-attribute ast 'image-2 default-image-name))

         (name-3 (ast-attribute ast 'name-3))
         (repeat-3  (as-boolean (ast-attribute ast 'repeat-3 "false")))
         (chord-match-3 (as-boolean (ast-attribute ast 'chord-match-3 "false")))
         (image-3 (ast-attribute ast 'image-3 default-image-name))

         (name-4 (ast-attribute ast 'name-4))
         (repeat-4  (as-boolean (ast-attribute ast 'repeat-4 "false")))
         (chord-match-4 (as-boolean (ast-attribute ast 'chord-match-4 "false")))
         (image-4 (ast-attribute ast 'image-4 default-image-name))

         (play-only-track-number (ast-attribute ast 'play-only #f))   ; play-only attribute on root construct
         (counter-transp (ast-attribute ast 'counterTransposition "0"))

         (midi-tracks (find-asts ast "MidiTrack"))
         (play-only-conditions
            (if play-only-track-number
                (play-only-track-number-to-play-only-condition play-only-track-number)
                (map find-play-only-condition midi-tracks)))

         (filtered-tracks 
            (cond ((equal? play-only-conditions (list #f #f #f #f))    ; no play only conditions specified
                     midi-tracks)
                  ((= 1 (count-true play-only-conditions))
                     (list (car (car (filter                        ; Tricky stuff: inner car: select from pair. Outer car: first element only.
                                 (lambda (tr-play)
                                   (let ((tr (car tr-play))
                                         (play (cdr tr-play)))
                                     play))
                                 (pair-up midi-tracks play-only-conditions))))))
                  (else (laml-error "Only a single track is allowed to be marked with 'play-only \"true\""))))

         (timing (if (equal? mode "absTime") (list 'absTime "0") (list 'deltaTime "0")))
        )
    (make-ast "StandardMidiFile"    ; Avoid calling (StandardMidiFile  ...), because it activates action procedure
      (cons
        (MidiHeader 'format "1" 'numberOfTracks "5" 'pulsesPerQuarterNote "1920" 'mode mode 'counterTransposition counter-transp)
        (cons
         (MidiTrack
          (Meta timing  'type "1" (multipad-repeat-string repeat-1 repeat-2 repeat-3 repeat-4))
          (Meta timing  'type "1" (multipad-chord-match-string chord-match-1 chord-match-2 chord-match-3 chord-match-4))
          (Meta timing  'type "1" (multipad-name "N1" name-1))
          (Meta timing  'type "1" (multipad-name "N2" name-2))
          (Meta timing  'type "1" (multipad-name "N3" name-3))
          (Meta timing  'type "1" (multipad-name "N4" name-4))
          (Meta timing  'type "1" (string-append "I1" image-1))
          (Meta timing  'type "1" (string-append "I2" image-2))
          (Meta timing  'type "1" (string-append "I3" image-3))
          (Meta timing  'type "1" (string-append "I4" image-4))
          (Meta timing  'type "47" "") ; end of track
          )
         filtered-tracks))
     '()  ; attributes
     'double
     "midi")))

(define (play-only-track-number-to-play-only-condition n)
  (let ((nn (as-number n)))
    (cond ((= nn 1) (list #t #f #f #f)) 
          ((= nn 2) (list #f #t #f #f))
          ((= nn 3) (list #f #f #t #f))
          ((= nn 4) (list #f #f #f #t))
          (else     (list #f #f #f #f))   ; all tracks
    )))


(define (find-play-only-condition track)
  (as-boolean (ast-attribute track 'play-only "false")))

(define (count-true bool-lst)
  (length (filter (lambda (x) x) bool-lst)))

(define (standard-midi-file-ast-to-multipad-ast ast input-file-path)
  (let* ((header (ast-subtree ast "MidiHeader"))
         (track-1 (ast-subtree ast "MidiTrack" 1))
         (remaining-tracks (cdr (find-asts ast "MidiTrack"))))
    (if (not (equal? "1" (ast-attribute header 'format))) (display-warning "Expecting a format 1 midi file"))
    (if (not (= 4 (length remaining-tracks))) (display-warning "Expecting 5 tracks in midi file"))
    (if (not (equal? "1920" (ast-attribute header 'pulsesPerQuarterNote) )) (display-warning "Expecting 1920 pulses per quarter note"))
;    (if (not (equal? "absTime" (ast-attribute header 'mode) )) (display-warning "Expecting mode absTime"))

    (let* ((meta-ast-list (find-asts track-1 "Meta"))
           (meta-ast-list-1 (filter (lambda (meta-ast) (equal? "1" (ast-attribute meta-ast 'type))) meta-ast-list))
           (meta-text-list (map ast-text meta-ast-list-1))
           (rp (find-meta "RP" meta-text-list))
           (cm (find-meta "CM" meta-text-list))
           (n1 (find-meta "N1" meta-text-list))
           (n2 (find-meta "N2" meta-text-list))
           (n3 (find-meta "N3" meta-text-list))
           (n4 (find-meta "N4" meta-text-list))
           (i1 (find-meta "I1" meta-text-list))
           (i2 (find-meta "I2" meta-text-list))
           (i3 (find-meta "I3" meta-text-list))
           (i4 (find-meta "I4" meta-text-list))
           
           (input-file-name-proper (file-name-proper input-file-path))   ; expected form x.S123.pad
           (image-file-name (file-name-extension input-file-name-proper))
           (default-image-file-name "S910")
          )
      (TyrosMultiPad
       'mode (ast-attribute header 'mode)
       'image (if (empty-string? image-file-name) default-image-file-name image-file-name)
       'name-1 (strip-trailing-null-chars n1)  
       'repeat-1 (if (eqv? (string-ref rp 0) #\0) "false" "true")
       'chord-match-1 (if (eqv? (string-ref cm 0) #\0) "false" "true")
       'image-1 i1

       'name-2 (strip-trailing-null-chars n2)  
       'repeat-2 (if (eqv? (string-ref rp 1) #\0) "false" "true")
       'chord-match-2 (if (eqv? (string-ref cm 1) #\0) "false" "true")
       'image-2 i2

       'name-3 (strip-trailing-null-chars n3)
       'repeat-3 (if (eqv? (string-ref rp 2) #\0) "false" "true")
       'chord-match-3 (if (eqv? (string-ref cm 2) #\0) "false" "true")
       'image-3 i3

       'name-4 (strip-trailing-null-chars n4)
       'repeat-4 (if (eqv? (string-ref rp 3) #\0) "false" "true")
       'chord-match-4 (if (eqv? (string-ref cm 3) #\0) "false" "true")
       'image-4 i4

       remaining-tracks))))


(define (find-meta prefix str-lst)
  (let ((res (find-in-list 
                (lambda (str) (equal? prefix (substring str 0 (string-length prefix))))
                str-lst)))
    (if res
        (substring res (string-length prefix) (string-length res))
        (laml-error "find-meta: Cannot find prefix" prefix "in" str-lst))))

(define (strip-trailing-null-chars str)
  (substring str 0 (find-in-string str #\nul)))


(define (multipad-repeat-string r1 r2 r3 r4)
  (string-append "RP"
    (if r1 "1" "0")    (if r2 "1" "0")     (if r3 "1" "0")     (if r4 "1" "0")))

(define (multipad-chord-match-string cm1 cm2 cm3 cm4)
  (string-append "CM"
    (if cm1 "1" "0")    (if cm2 "1" "0")     (if cm3 "1" "0")     (if cm4 "1" "0")))

 
(define (multipad-name prefix name)
 (let ((padding-length (- 52 (+ (string-length prefix) (string-length name)))))
  (string-append prefix name (make-string padding-length #\nul))))

; ---------------------------------------------------------------------------------------------------------------
; Chords

; A list of all chord formula - taken from the Yahama Tyros data list (page 23).
(define chord-types
  '(
    ("M"        "major"                                0 4 7)                ; 0
    ("6"        "sixth"                                0 4 7 9)              ; 1
    ("M7"       "major seventh"                        0 4 7 11)             ; 2
    ("M7(#11)"  "major seventh add sharp eleventh"     0 2 4 6 7 11)         ; 3
    ("9"        "add ninth"                            0 2 4 7)              ; 4
    ("M7_9"     "major seventh ninth"                  0 2 4 7 11)           ; 5
    ("6_9"      "sixht ninth"                          0 2 4 7 9)            ; 6
    ("aug"      "augmented"                            0 4 8)                ; 7
  
    ("m"        "minor"                                0 3 7)                ; 8
    ("m6"       "minor sixth"                          0 3 7 9)              ; 9
    ("m7"       "minor seventh"                        0 3 7 10)             ; 10
    ("m7b5"     "minor seventh flat fifth"             0 3 6 10)             ; 11
    ("m(9)"     "minor add ninth"                      0 2 3 7)              ; 12
    ("m7(9)"    "minor seventh ninth"                  0 2 3 7 10)           ; 13
    ("m7(11)"   "minor seventh eleventh"               0 2 3 5 7 10)         ; 14
    ("mM7"      "minor major seventh"                  0 3 7 11)             ; 15
    ("mM7(9)"   "minor major seventh ninth"            0 2 3 7 11)           ; 16
  
    ("dim"       "diminished"                          0 3 6)                ; 17
    ("dim7"      "diminished seventh"                  0 3 6 9)              ; 18
    ("7"         "seventh"                             0 4 7 10)             ; 19
    ("7sus4"     "seventh suspended fourth"            0 5 7 10)             ; 20
    ("7b5"       "sevent flatted fifth"                0 4 6 10)             ; 21
    ("7(9)"      "seventh ninth"                       0 2 4 7 10)           ; 22
    ("7(#11)"    "seventh add sharp eleventh"          0 2 4 6 7 10)         ; 23
    ("7(13)"     "seventh add thirteenth"              0 4 7 9 10)           ; 24
    ("7(b9)"     "sevent flatted ninth"                0 1 4 7 10)           ; 25
    ("7(b13)"    "seventh add flatted thirteenth"      0 4 7 8 10)           ; 26
    ("7(#9)"     "seventh sharp ninth"                 0 3 4 7 10)           ; 27
    ("M7aug"    "major seventh augmented"              0 4 8 11)             ; 28
    ("7aug"     "seventh augmented"                    0 4 8 10)             ; 29
    ("1+8"      "1+8"                                  0 1 3)                ; 30
    ("1+5"      "1+5"                                  0 7)                  ; 31
    ("sus4"      "suspended fourth"                    0 5 7)                ; 32
    ("1+2+5"     "???"                                 0 2 5)                ; 33   ; wrong!
    ("cancel"    "cancel"                              0 1 2)                ; 34

    ("sus2"      "suspended plus two plus five"        0 2 7)                ; 35   ; non-yamaha
    ("M7b5"     "major seventh flat fifth"             0 4 6 11)             ; 36   ; non-yamaha
    ("b5"       "flatted fifth"                        0 4 6)                ; 37   ; non-yamaha
    ("mM7b5"    "minor major seventh flatted fifth"    0 3 6 11)             ; 38   ; non-yamaha
    
  ))

; Return the chord given type. 
; This function accept both short chord name (a string) and a number
; The numering correspond the Yamaha/Tyros ordering.
(define (lookup-chord-type-formula type)
  (cond ((string? type)
          (let ((res (assoc type chord-types)))
            (if res
                (cddr res)
                (laml-error "lookup-chord-type-formula: Unknown chord type:" type))))

        ((integer? type)
          (cddr (list-ref chord-types type)))

        (else (laml-error "lookup-chord-type-formula: Input must be string (short chord name) or an index number (between 0 and 38)"))))

(define (lookup-chord-type-short-name chord-number)
  (if (and (>= chord-number 0) (<= chord-number 38))
      (let ((res (list-ref chord-types chord-number)))
        (car res))
      "??"))
  
      
  


; Return a chord list for a given chord.
; Root is a name (string): C, C#, D, D#, E, F, F#, G, G#, A, A#, B.
; Chord-type is a chord type name, as appearing in the list chord-types.
; start-octave is an octave number. Octave number 3 contains the middle C (note 60) - according the Yamaha convention. Within the interval [-2..8]
; Generate it through number-of-octaves (a non-negative integer), allways ended by the root note.
; .form (chord-note-list root chord-type start-octave number-of-octaves)
(define (chord-note-list root chord-type start-octave number-of-octaves)
  (let* ((relative-chord-formula (lookup-chord-type-formula (as-string chord-type)))
         (root-number (number-of-note-name (as-string root)))
         (relative-root-chord-formula (map (lambda (n) (+ n root-number)) relative-chord-formula))
         (relative-repeated-root-chord-formula (repeat-chord-formula relative-root-chord-formula number-of-octaves))
         (start-octave-number (start-of-octave-number start-octave))
         (absolute-repeated-root-chord-formula (map (lambda (n) (+ n start-octave-number)) relative-repeated-root-chord-formula))
        )
    absolute-repeated-root-chord-formula 
  ))

; Return a chord list (a list of integers) of length lgt, ending in note, played according to chord-type with a given root.
; Notice that  note  always will be the last element of the list (if a list is returned, of course).
; If it is not possible to form a chor note list of the given chord-type and root, return #f.
(define (chord-note-list-ending-at note lgt root chord-type)
  (let* ((total-long-chord-list (chord-note-list root chord-type -2 10))   ; all notes in given chord-type with given root
         (idx-of-note (index-in-list-by-predicate total-long-chord-list note =)))
    (if idx-of-note
        (let* ((to-idx (+ idx-of-note 1))
               (from-idx (- to-idx (- lgt 1)))) 
          (list-part from-idx to-idx total-long-chord-list))
        #f)))

(define (repeat-chord-formula formula n)
  (if (= n 0) 
      (list (car formula))   ; ending at root
      (append formula 
              (repeat-chord-formula 
               (map (lambda (n) (+ n 12)) formula) (- n 1)))))

(define (start-of-octave-number n)
  (cond 
        ((< n -2) (laml-error "start-of-octave-number: Octave numbers must be within [-2..8]" n))
        ((= n -2) 0)
        ((= n -1) 12)
        ((= n 0) 24)
        ((= n 1) 36)
        ((= n 2) 48)
        ((= n 3) 60)
        ((= n 4) 72)
        ((= n 5) 84)
        ((= n 6) 96)
        ((= n 7) 108)
        ((= n 8) 120)
        ((> n 8) (laml-error "start-of-octave-number: Octave numbers must be within [-2..8]" n))))

    
; ---------------------------------------------------------------------------------------------------------------
; Track separation.
; Transformation of MIDI format 0 AST to MIDI format 1 AST.

; Defines how to separate tracks when making a format 1 AST.
(define standard-track-separation-scheme 
  '((system meta) 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16))

; Create a format 1 ast from a format 0 StandardMidiFile ast
(define (track-separated-standard-midi-file-ast format-0-ast track-separation-scheme)
  (make-ast "StandardMidiFile"    ; use make-ast to avoid calling the mirror function StandardMidiFile. We do not wish to perfom the action procedure.
    (cons 
      (MidiHeader 'format "1" 
                  'numberOfTracks (as-string (length track-separation-scheme)) 
                  'pulsesPerQuarterNote (unique-ast-attribute format-0-ast 'pulsesPerQuarterNote)
                  'mode (unique-ast-attribute format-0-ast 'mode)
                  'counterTransposition (unique-ast-attribute format-0-ast 'counterTransposition "0"))
      (make-multiple-level-1-tracks-from-single-track (ast-subtree format-0-ast "MidiTrack")  ; only one such "MidiTrack" AST in format 0 StandardMidiFile AST
                            track-separation-scheme)
    )
    '()  ; attributes
    'double
    "midi"))

(define (make-multiple-level-1-tracks-from-single-track midi-track-ast track-separation-scheme)
  (map 
    (lambda (track-separation-item)
      (let ((event-ast-list (ast-subtrees midi-track-ast)))
        (MidiTrack (filter (track-predicate track-separation-item) event-ast-list))))
    track-separation-scheme))

(define system-meta-element-names
  (list "SysEx" "MTCQuarterFrame" "SongPositionPointer" "SongSelect" "TuneRequest" "TimingClock" "Start" "Continue" "Stop" "ActiveSensing"))

; Return an AST predicate that recognizes an item from the track-separation-scheme  (an element in that list).
; Legal track-items:   (system meta), (meta system), meta, system, an integer, a non-empty integer list. 
(define (track-predicate track-item) 
  (cond ((and (pair? track-item) (or (equal? track-item '(system meta))  (equal? track-item '(meta system))))
           (lambda (x)
              (and (ast? x) (member (ast-element-name x) (cons "Meta" system-meta-element-names)))))
        ((int-list? track-item)
           (lambda (x)
              (and (ast? x) (member (as-number (ast-attribute x 'channel)) track-item))))
        ((and (symbol? track-item) (equal? track-item 'system))
           (lambda (x)
              (and (ast? x) (member (ast-element-name x) system-meta-element-names))))
        ((and (symbol? track-item) (equal? track-item 'meta))
           (lambda (x)
              (and (ast? x) (equal? (ast-element-name x) "Meta"))))
        ((integer? track-item)
           (lambda (x)
              (and (ast? x) (= track-item (as-number (ast-attribute x 'channel))))))
        (else
           (lambda (x)
               #f))))

(define (int-list? x)
  (and (pair? x) (list? x) (>= (length x) 1)
       (accumulate-right and-fn #t (map integer? x))))
       
; ---------------------------------------------------------------------------------------------------------------   
; Track joining
; Transformation of MIDI format 1 AST to MIDI format 0 AST.

(define (track-joined-standard-midi-file-ast format-1-ast)
  (make-ast "StandardMidiFile"    ; use make-ast to avoid calling the mirror function StandardMidiFile. We do not wish to perfom the action procedure.
    (cons 
      (MidiHeader 'format "0" 
                  'numberOfTracks "1" 
                  'pulsesPerQuarterNote (unique-ast-attribute format-1-ast 'pulsesPerQuarterNote)
                  'mode (unique-ast-attribute format-1-ast 'mode)
                  'counterTransposition (unique-ast-attribute format-1-ast 'counterTransposition "0"))
      (cons #t (list 
                (make-single-track-from-multiple-tracks 
                  (filter 
                    (lambda (x) (and (ast? x) (equal? "MidiTrack" (ast-element-name x))))
                    (ast-subtrees format-1-ast))
                ))
      )
    )
    '()  ; attributes
    'double
    "midi"))

(define (make-single-track-from-multiple-tracks midi-track-ast-list)
 (MidiTrack
  (stable-sort-list 
    (flatten
     (map 
       (lambda (midi-track-ast)
          (get-rid-of-delta-times (filter ast? (ast-subtrees midi-track-ast)) 0))
       midi-track-ast-list))
    abs-time-ast-leq?))
)

; Consider a more refined function. As lo
(define (abs-time-ast-leq? ast1 ast2)
  (<= (as-number (ast-attribute ast1 'absTime)) (as-number (ast-attribute ast2 'absTime))))

; A list sorting function which is stable.
(define (stable-sort-list lst compare-fn)
  (sort-list (reverse lst) compare-fn))

; ---------------------------------------------------------------------------------------------------------------

(define (meta-division-event? x)
  (and x
       (ast? x)
       (equal? (ast-element-name x) "Meta")
       (equal? "6" (ast-attribute x 'type))))

(define (no-spaces-in-string str)
  (transliterate str #\space "-"))

; Return a consequtive subset of the events in track-events.
; from and to are names (strings) of type 6 meta events (markers).
; from and to can also be booleans. 
; If from is #t, it means from beginning.
; If to is #t, it means to end.
(define (midi-event-ast-subsequence track-events from to)
  (sublist-of-list 
    track-events
    (if (and (boolean? from) from) 
        (lambda (x) #t)
        (lambda (x)
          (and
           (ast? x)
           (equal? (ast-element-name x) "Meta")
           (equal? "6" (ast-attribute x 'type))
           (equal? from (ast-text x)))))
    (if (and (boolean? to) to) 
        (lambda (x) #f)
        (lambda (x)
          (and
           (ast? x)
           (equal? (ast-element-name x) "Meta")
           (equal? "6" (ast-attribute x 'type))
           (equal? to (ast-text x)))))))                   
